import keras
from keras import ops

import numpy as np

import warnings

from bayesflow.networks import MLP
from bayesflow.types import Tensor
from bayesflow.utils import (
    jvp,
    concatenate_valid,
    find_network,
    expand_right_as,
    expand_right_to,
    model_kwargs,
)
from bayesflow.utils.serialization import deserialize, serializable, serialize


from bayesflow.networks import InferenceNetwork
from bayesflow.networks.embeddings import FourierEmbedding


@serializable
class ContinuousTimeConsistencyModel(InferenceNetwork):
    """Implements an sCM (simple, stable, and scalable Consistency Model)
    with continous-time Consistency Training (CT) as described in [1].
    The sampling procedure is taken from [2].

    [1] Lu, C., & Song, Y. (2024).
    Simplifying, Stabilizing and Scaling Continuous-Time Consistency Models
    arXiv preprint arXiv:2410.11081

    [2] Song, Y., Dhariwal, P., Chen, M. & Sutskever, I. (2023).
    Consistency Models.
    arXiv preprint arXiv:2303.01469
    """

    def __init__(
        self,
        subnet: str | keras.Layer = "mlp",
        sigma_data: float = 1.0,
        subnet_kwargs: dict[str, any] = None,
        embedding_kwargs: dict[str, any] = None,
        **kwargs,
    ):
        """Creates an instance of an sCM to be used for consistency training (CT).

        Parameters
        ----------
        subnet        : str or type, optional, default: "mlp"
            A neural network type for the consistency model, will be
            instantiated using subnet_kwargs.
        sigma_data    : float, optional, default: 1.0
            Standard deviation of the target distribution
        **kwargs
            Additional keyword arguments to the layer.
        """
        super().__init__(base_distribution="normal", **kwargs)

        if subnet_kwargs:
            warnings.warn(
                "Using `subnet_kwargs` is deprecated."
                "Instead, instantiate the network yourself and pass the arguments directly.",
                DeprecationWarning,
            )

        subnet_kwargs = subnet_kwargs or {}

        self.subnet = find_network(subnet, **subnet_kwargs)
        self.subnet_projector = keras.layers.Dense(
            units=None, bias_initializer="zeros", kernel_initializer="zeros", name="subnet_projector"
        )

        self.weight_fn = MLP([256], dropout=0.0)
        self.weight_fn_projector = keras.layers.Dense(
            units=1, bias_initializer="zeros", kernel_initializer="zeros", name="weight_fn_projector"
        )

        embedding_kwargs = embedding_kwargs or {}
        self.time_emb = FourierEmbedding(**embedding_kwargs)
        self.time_emb_dim = self.time_emb.embed_dim

        self.sigma_data = sigma_data

        self.seed_generator = keras.random.SeedGenerator()

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        base_config = super().get_config()
        base_config = model_kwargs(base_config)

        config = {
            "subnet": self.subnet,
            "sigma_data": self.sigma_data,
        }

        return base_config | serialize(config)

    def _discretize_time(self, num_steps: int, rho: float = 3.5, **kwargs):
        t = np.linspace(0.0, np.pi / 2, num_steps)
        times = np.exp((t - np.pi / 2) * rho) * np.pi / 2
        times[0] = 0.0

        # if rho is set too low, bad schedules can occur
        EPS_WARN = 0.1
        if times[1] > EPS_WARN:
            print("Warning: The last time step is large.")
            print(f"Increasing rho (was {rho}) or n_steps (was {num_steps}) might improve results.")
        return ops.convert_to_tensor(times)

    def build(self, xz_shape, conditions_shape=None):
        super().build(xz_shape)
        self.subnet_projector.units = xz_shape[-1]

        # construct input shape for subnet and subnet projector
        input_shape = list(xz_shape)

        # time vector
        input_shape[-1] += self.time_emb_dim + 1

        if conditions_shape is not None:
            input_shape[-1] += conditions_shape[-1]

        input_shape = tuple(input_shape)

        self.subnet.build(input_shape)

        input_shape = self.subnet.compute_output_shape(input_shape)
        self.subnet_projector.build(input_shape)

        # input shape for time embedding
        self.time_emb.build((xz_shape[0], 1))

        # input shape for weight function and projector
        input_shape = (xz_shape[0], 1)
        self.weight_fn.build(input_shape)
        input_shape = self.weight_fn.compute_output_shape(input_shape)
        self.weight_fn_projector.build(input_shape)

    def call(
        self,
        xz: Tensor,
        conditions: Tensor = None,
        inverse: bool = False,
        **kwargs,
    ):
        if inverse:
            return self._inverse(xz, conditions=conditions, **kwargs)
        return self._forward(xz, conditions=conditions, **kwargs)

    def _forward(self, x: Tensor, conditions: Tensor = None, **kwargs) -> Tensor:
        # Consistency Models only learn the direction from noise distribution
        # to target distribution, so we cannot implement this function.
        raise NotImplementedError("Consistency Models are not invertible")

    def _inverse(self, z: Tensor, conditions: Tensor = None, **kwargs) -> Tensor:
        """Generate random draws from the approximate target distribution
        using the multistep sampling algorithm from [2], Algorithm 1.

        Parameters
        ----------
        z           : Tensor
            Samples from a standard normal distribution
        conditions  : Tensor, optional, default: None
            Conditions for a approximate conditional distribution
        **kwargs    : dict, optional, default: {}
            Additional keyword arguments. Include `steps` (default: 30) to
            adjust the number of sampling steps.

        Returns
        -------
        x            : Tensor
            The approximate samples
        """
        steps = kwargs.get("steps", 15)
        rho = kwargs.get("rho", 3.5)

        # noise distribution has variance sigma_data
        x = keras.ops.copy(z) * self.sigma_data
        discretized_time = keras.ops.flip(self._discretize_time(steps, rho=rho), axis=-1)
        t = keras.ops.full((*keras.ops.shape(x)[:-1], 1), discretized_time[0], dtype=x.dtype)
        x = self.consistency_function(x, t, conditions=conditions)
        for n in range(1, steps):
            noise = keras.random.normal(keras.ops.shape(x), dtype=keras.ops.dtype(x), seed=self.seed_generator)
            x_n = ops.cos(t) * x + ops.sin(t) * noise
            t = keras.ops.full_like(t, discretized_time[n])
            x = self.consistency_function(x_n, t, conditions=conditions)
        return x

    def consistency_function(
        self,
        x: Tensor,
        t: Tensor,
        conditions: Tensor = None,
        training: bool = False,
        **kwargs,
    ) -> Tensor:
        """Compute consistency function at time t.

        Parameters
        ----------
        x           : Tensor
            Input vector
        t           : Tensor
            Vector of time samples in [0, pi/2]
        conditions  : Tensor
            The conditioning vector
        training    : bool
            Flag to control whether the inner network operates in training or test mode
        **kwargs    : dict, optional, default: {}
            Additional keyword arguments passed to the inner network.
        """
        xtc = concatenate_valid([x / self.sigma_data, self.time_emb(t), conditions], axis=-1)
        f = self.subnet_projector(self.subnet(xtc, training=training, **kwargs))
        out = ops.cos(t) * x - ops.sin(t) * self.sigma_data * f
        return out

    def compute_metrics(
        self, x: Tensor, conditions: Tensor = None, stage: str = "training", **kwargs
    ) -> dict[str, Tensor]:
        base_metrics = super().compute_metrics(x, conditions=conditions, stage=stage)

        # $# Implements Algorithm 1 from [1]

        # training parameters
        p_mean = -1.0
        p_std = 1.6

        c = 0.1

        # generate noise vector
        z = (
            keras.random.normal(keras.ops.shape(x), dtype=keras.ops.dtype(x), seed=self.seed_generator)
            * self.sigma_data
        )

        # sample time
        tau = (
            keras.random.normal(keras.ops.shape(x)[:1], dtype=keras.ops.dtype(x), seed=self.seed_generator) * p_std
            + p_mean
        )
        t_ = ops.arctan(ops.exp(tau) / self.sigma_data)
        t = expand_right_as(t_, x)

        # generate noisy sample
        xt = ops.cos(t) * x + ops.sin(t) * z

        # calculate estimator for dx_t/dt
        dxtdt = ops.cos(t) * z - ops.sin(t) * x

        r = 1.0  # TODO: if consistency distillation training (not supported yet) is unstable, add schedule here

        def f_teacher(x, t):
            o = self.subnet(concatenate_valid([x, self.time_emb(t), conditions], axis=-1), training=stage == "training")
            return self.subnet_projector(o)

        primals = (xt / self.sigma_data, t)
        tangents = (
            ops.cos(t) * ops.sin(t) * dxtdt,
            ops.cos(t) * ops.sin(t) * self.sigma_data,
        )

        teacher_output, cos_sin_dFdt = jvp(f_teacher, primals, tangents, return_output=True)
        teacher_output = ops.stop_gradient(teacher_output)
        cos_sin_dFdt = ops.stop_gradient(cos_sin_dFdt)

        # calculate output of the network
        xtc = concatenate_valid([xt / self.sigma_data, self.time_emb(t), conditions], axis=-1)
        student_out = self.subnet_projector(self.subnet(xtc, training=stage == "training"))

        # calculate the tangent
        g = -(ops.cos(t) ** 2) * (self.sigma_data * teacher_output - dxtdt) - r * ops.cos(t) * ops.sin(t) * (
            xt + self.sigma_data * cos_sin_dFdt
        )

        # apply normalization to stabilize training
        g = g / (ops.norm(g, axis=-1, keepdims=True) + c)

        # compute adaptive weights and calculate loss
        w = self.weight_fn_projector(self.weight_fn(expand_right_to(t_, 2)))

        D = ops.shape(x)[-1]
        loss = ops.mean(
            (ops.exp(w) / D)
            * ops.mean(
                ops.reshape(((student_out - teacher_output - g) ** 2), (ops.shape(teacher_output)[0], -1)), axis=-1
            )
            - w
        )

        return base_metrics | {"loss": loss}



"""
A collection of `keras.utils.PyDataset <https://keras.io/api/utils/python_utils/#pydataset-class>`__\ s, which
wrap your data-generating process (i.e., your :py:class:`~bayesflow.simulators.Simulator`) and thus determine the
effective training strategy (e.g., online or offline).
"""

from .offline_dataset import OfflineDataset
from .online_dataset import OnlineDataset
from .disk_dataset import DiskDataset
from .rounds_dataset import RoundsDataset

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])



import keras
import pytest

from bayesflow.utils import find_inference_network, find_distribution, find_summary_network


# --- Tests for find_inference_network.py ---


class DummyInferenceNetwork:
    def __init__(self, *a, **kw):
        self.args = a
        self.kwargs = kw


@pytest.mark.parametrize(
    "name,expected_class_path",
    [
        ("coupling_flow", "bayesflow.networks.CouplingFlow"),
        ("flow_matching", "bayesflow.networks.FlowMatching"),
        ("consistency_model", "bayesflow.networks.ConsistencyModel"),
    ],
)
def test_find_inference_network_by_name(monkeypatch, name, expected_class_path):
    # patch the expected class in bayesflow.networks
    components = expected_class_path.split(".")
    module_path = ".".join(components[:-1])
    class_name = components[-1]

    dummy_cls = DummyInferenceNetwork
    monkeypatch.setattr(f"{module_path}.{class_name}", dummy_cls)

    net = find_inference_network(name, 1, key="val")
    assert isinstance(net, DummyInferenceNetwork)
    assert net.args == (1,)
    assert net.kwargs == {"key": "val"}


def test_find_inference_network_by_keras_layer():
    layer = keras.layers.Dense(10)
    result = find_inference_network(layer)
    assert result is layer


def test_find_inference_network_by_keras_model():
    model = keras.models.Sequential()
    result = find_inference_network(model)
    assert result is model


def test_find_inference_network_unknown_name():
    with pytest.raises(ValueError):
        find_inference_network("unknown_network_name")


def test_find_inference_network_invalid_type():
    with pytest.raises(TypeError):
        find_inference_network(12345)


# --- Tests for find_distribution.py ---


class DummyDistribution:
    def __init__(self, *a, **kw):
        self.args = a
        self.kwargs = kw


@pytest.mark.parametrize(
    "name, expected_class_path",
    [
        ("normal", "bayesflow.distributions.DiagonalNormal"),
        ("student", "bayesflow.distributions.DiagonalStudentT"),
        ("student-t", "bayesflow.distributions.DiagonalStudentT"),
        ("student_t", "bayesflow.distributions.DiagonalStudentT"),
    ],
)
def test_find_distribution_by_name(monkeypatch, name, expected_class_path):
    components = expected_class_path.split(".")
    module_path = ".".join(components[:-1])
    class_name = components[-1]

    dummy_cls = DummyDistribution
    monkeypatch.setattr(f"{module_path}.{class_name}", dummy_cls)

    dist = find_distribution(name, 10, a=5)
    assert isinstance(dist, DummyDistribution)
    assert dist.args == (10,)
    assert dist.kwargs == {"a": 5}


def test_find_distribution_none_returns_none():
    assert find_distribution(None) is None


def test_find_distribution_with_keras_layer():
    layer = keras.layers.Dense(3)
    result = find_distribution(layer)
    assert result is layer


def test_find_distribution_mixture_raises():
    with pytest.raises(ValueError):
        find_distribution("mixture")


def test_find_distribution_invalid_name():
    with pytest.raises(ValueError):
        find_distribution("invalid_name")


def test_find_distribution_invalid_type():
    with pytest.raises(TypeError):
        find_distribution(3.14)


# --- Tests for find_summary_network.py ---


class DummySummaryNetwork:
    def __init__(self, *a, **kw):
        self.args = a
        self.kwargs = kw


@pytest.mark.parametrize(
    "name,expected_class_path",
    [
        ("deep_set", "bayesflow.networks.DeepSet"),
        ("set_transformer", "bayesflow.networks.SetTransformer"),
        ("fusion_transformer", "bayesflow.networks.FusionTransformer"),
        ("time_series_transformer", "bayesflow.networks.TimeSeriesTransformer"),
        ("time_series_network", "bayesflow.networks.TimeSeriesNetwork"),
    ],
)
def test_find_summary_network_by_name(monkeypatch, name, expected_class_path):
    components = expected_class_path.split(".")
    module_path = ".".join(components[:-1])
    class_name = components[-1]

    dummy_cls = DummySummaryNetwork
    monkeypatch.setattr(f"{module_path}.{class_name}", dummy_cls)

    net = find_summary_network(name, 22, flag=True)
    assert isinstance(net, DummySummaryNetwork)
    assert net.args == (22,)
    assert net.kwargs == {"flag": True}


def test_find_summary_network_by_keras_layer():
    layer = keras.layers.Dense(1)
    out = find_summary_network(layer)
    assert out is layer


def test_find_summary_network_by_keras_model():
    model = keras.models.Sequential()
    out = find_summary_network(model)
    assert out is model


def test_find_summary_network_unknown_name():
    with pytest.raises(ValueError):
        find_summary_network("unknown_summary_net")


def test_find_summary_network_invalid_type():
    with pytest.raises(TypeError):
        find_summary_network(0.1234)



import keras

from bayesflow.types import Tensor
from bayesflow.utils.serialization import serializable

from .skip_recurrent import SkipRecurrentNet
from ..summary_network import SummaryNetwork


@serializable
class TimeSeriesNetwork(SummaryNetwork):
    """
    Implements a LSTNet Architecture as described in [1]

    [1] Y. Zhang and L. Mikelsons, Solving Stochastic Inverse Problems with Stochastic BayesFlow,
    2023 IEEE/ASME International Conference on Advanced Intelligent Mechatronics (AIM),
    Seattle, WA, USA, 2023, pp. 966-972, doi: 10.1109/AIM46323.2023.10196190.
    """

    def __init__(
        self,
        summary_dim: int = 16,
        filters: int | list | tuple = 32,
        kernel_sizes: int | list | tuple = 3,
        strides: int | list | tuple = 1,
        activation: str = "mish",
        kernel_initializer: str = "glorot_uniform",
        groups: int = None,
        recurrent_type: str = "gru",
        recurrent_dim: int = 128,
        bidirectional: bool = True,
        dropout: float = 0.05,
        skip_steps: int = 4,
        **kwargs,
    ):
        """
        Initializes a hybrid convolutional-recurrent network for summarizing multivariate time series.

        This model combines a convolutional backbone with a recurrent module to efficiently process sequential data.

        The convolutional layers extract local features, which are then processed by a recurrent network for capturing
        long-range dependencies. The model supports various activation functions, kernel initializations, and
        bidirectional recurrence for enhanced representation learning.

        Parameters
        ----------
        summary_dim : int, optional
            Dimensionality of the final summary representation. Default is 16.
        filters : int, list, or tuple, optional
            Number of filters in each convolutional layer. If an integer is provided,
            all layers will have the same number of filters. Default is 32.
        kernel_sizes : int, list, or tuple, optional
            Size of the convolutional kernels. If an integer is provided, all layers
            will have the same kernel size. Default is 3.
        strides : int, list, or tuple, optional
            Stride length for convolutional layers. If an integer is provided, all layers
            will have the same stride. Default is 1.
        activation : str, optional
            Activation function applied in the convolutional layers. Default is "mish".
        kernel_initializer : str, optional
            Initialization strategy for convolutional kernels, such as "glorot_uniform".
            Default is "glorot_uniform".
        groups : int, optional
            Number of groups for group normalization applied after each convolutional layer.
            Default is None.
        recurrent_type : str, optional
            Type of recurrent layer used for sequence modeling, such as "gru" or "lstm".
            Default is "gru".
        recurrent_dim : int, optional
            Number of hidden units in the recurrent layer. Default is 128.
        bidirectional : bool, optional
            Whether to use a bidirectional recurrent network. Default is True.
        dropout : float, optional
            Dropout rate applied in the recurrent module. Default is 0.05.
        skip_steps : int, optional
            Number of steps to skip in the recurrent network for efficiency. Default is 4.
        **kwargs
            Additional keyword arguments passed to the parent class.
        """

        super().__init__(**kwargs)

        # Convolutional backbone -> can be extended with inception-like structure
        if not isinstance(filters, (list, tuple)):
            filters = (filters,)
        if not isinstance(kernel_sizes, (list, tuple)):
            kernel_sizes = (kernel_sizes,)
        if not isinstance(strides, (list, tuple)):
            strides = (strides,)

        conv_blocks = []
        for i, (f, k, s) in enumerate(zip(filters, kernel_sizes, strides)):
            conv_blocks.append(
                keras.layers.Conv1D(
                    filters=f,
                    kernel_size=k,
                    strides=s,
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                    padding="same",
                    name=f"conv_{i}",
                )
            )
            if groups is not None:
                conv_blocks.append(keras.layers.GroupNormalization(groups=groups, name=f"group_norm_{i}"))

        self.conv = keras.Sequential(conv_blocks, name="conv")

        # Recurrent and feedforward backbones
        self.recurrent = SkipRecurrentNet(
            hidden_dim=recurrent_dim,
            recurrent_type=recurrent_type,
            bidirectional=bidirectional,
            input_channels=filters[-1],
            skip_steps=skip_steps,
            dropout=dropout,
            name="recurrent",
        )
        self.output_projector = keras.layers.Dense(summary_dim, name="output_projector")

        self.summary_dim = summary_dim

    def call(self, x: Tensor, training: bool = False, **kwargs) -> Tensor:
        """
        Performs the forward pass of the hybrid convolutional-recurrent network.

        This function applies a sequence of convolutional layers followed by a recurrent module to extract spatial
        and temporal features from the input tensor.

        The final output is projected into a lower-dimensional summary representation using a dense layer.

        Parameters
        ----------
        x : Tensor
            Input tensor representing the sequence data.
        training : bool, optional
            Whether the model is in training mode, affecting layers like dropout and
            batch normalization. Default is False.
        **kwargs
            Additional keyword arguments passed to the layers.

        Returns
        -------
        output: Tensor
            Transformed tensor representing the summarized feature representation
            of the input sequence.
        """
        x = self.conv(x, training=training)
        x = self.recurrent(x, training=training)
        x = self.output_projector(x)
        return x



from typing import Sequence

import numpy as np


def confusion_matrix(targets: np.ndarray, estimates: np.ndarray, labels: Sequence = None, normalize: str = None):
    """
    Compute confusion matrix to evaluate the accuracy of a classification or model comparison setting.

    Code inspired by: https://github.com/scikit-learn/scikit-learn/blob/98ed9dc73/sklearn/metrics/_classification.py

    Parameters
    ----------
    targets : np.ndarray
        Ground truth (correct) target values.
    estimates : np.ndarray
        Estimated targets as returned by a classifier.
    labels : Sequence, optional
        List of labels to index the matrix. This may be used to reorder or select a subset of labels.
        If None, labels that appear at least once in y_true or y_pred are used in sorted order.
    normalize : {'true', 'pred', 'all'}, optional
        Normalizes confusion matrix over the true (rows), predicted (columns)
        conditions or all the population. If None, no normalization is applied.

    Returns
    -------
    cm : np.ndarray of shape (num_labels, num_labels)
        Confusion matrix. Rows represent true classes, columns represent predicted classes.
    """

    # Get unique labels
    if labels is None:
        labels = np.unique(np.concatenate((targets, estimates)))
    else:
        labels = np.asarray(labels)

    label_to_index = {label: i for i, label in enumerate(labels)}
    num_labels = len(labels)

    # Initialize the confusion matrix
    cm = np.zeros((num_labels, num_labels), dtype=np.int64)

    # Fill confusion matrix
    for t, p in zip(targets, estimates):
        if t in label_to_index and p in label_to_index:
            cm[label_to_index[t], label_to_index[p]] += 1

    # Normalize if required
    if normalize == "true":
        with np.errstate(all="ignore"):
            cm = cm.astype(np.float64)
            cm = np.divide(cm, cm.sum(axis=1, keepdims=True), where=cm.sum(axis=1, keepdims=True) != 0)
    elif normalize == "pred":
        with np.errstate(all="ignore"):
            cm = cm.astype(np.float64)
            cm = np.divide(cm, cm.sum(axis=0, keepdims=True), where=cm.sum(axis=0, keepdims=True) != 0)
    elif normalize == "all":
        cm = cm.astype(np.float64)
        cm /= cm.sum()

    return cm



import keras
import numpy as np

from bayesflow.types import Tensor
from bayesflow.utils import pad, searchsorted
from bayesflow.utils.keras_utils import shifted_softplus
from bayesflow.utils.serialization import serializable

from ._rational_quadratic import _rational_quadratic_spline
from .transform import Transform


@serializable
class SplineTransform(Transform):
    def __init__(
        self,
        bins: int = 16,
        default_domain: (float, float, float, float) = (-3.0, 3.0, -3.0, 3.0),
        min_width: float = 1.0,
        min_height: float = 1.0,
        min_bin_width: float = 0.1,
        min_bin_height: float = 0.1,
        method: str = "rational_quadratic",
    ):
        super().__init__()

        if bins <= 0:
            raise ValueError("Number of bins must be strictly positive.")

        self.bins = bins
        self.min_width = max(min_width, bins * min_bin_width)
        self.min_height = max(min_height, bins * min_bin_height)
        self.min_bin_width = min_bin_width
        self.min_bin_height = min_bin_height
        self.method = method

        if self.method != "rational_quadratic":
            raise NotImplementedError("Currently, only 'rational_quadratic' spline method is supported.")

        self.method_fn = _rational_quadratic_spline

        # we slightly over-parametrize to allow for better constraints
        # this may also improve convergence due to redundancy
        self.parameter_sizes = {
            "left_edge": 1,
            "bottom_edge": 1,
            "total_width": 1,
            "total_height": 1,
            "bin_widths": self.bins,
            "bin_heights": self.bins,
            "derivatives": self.bins - 1,
        }

        if default_domain[1] <= default_domain[0] or default_domain[3] <= default_domain[2]:
            raise ValueError("Invalid default domain. Must be (left, right, bottom, top).")

        self.default_left = default_domain[0]
        self.default_bottom = default_domain[2]
        self.default_width = default_domain[1] - default_domain[0]
        self.default_height = default_domain[3] - default_domain[2]

        if self.default_width < self.min_width:
            raise ValueError(f"Default width must be greater than minimum width ({self.min_width}).")

        if self.default_height < self.min_height:
            raise ValueError(f"Default height must be greater than minimum height ({self.min_height}).")

        self._shift = np.sinh(1.0) * np.log(np.e - 1.0)

    @property
    def params_per_dim(self) -> int:
        return sum(self.parameter_sizes.values())

    def split_parameters(self, parameters: Tensor) -> dict[str, Tensor]:
        batch_shape = list(keras.ops.shape(parameters)[:-1])
        parameters = keras.ops.reshape(parameters, batch_shape + [-1, self.params_per_dim])
        indices = np.cumsum(list(self.parameter_sizes.values())).tolist()
        parameters = keras.ops.split(parameters, indices, axis=-1)
        parameters = dict(zip(self.parameter_sizes.keys(), parameters))

        return parameters

    def constrain_parameters(self, parameters: dict[str, Tensor]) -> dict[str, Tensor]:
        left_edge = parameters["left_edge"] + self.default_left
        bottom_edge = parameters["bottom_edge"] + self.default_bottom

        # strictly positive (softplus)
        # scales logarithmically to infinity (arcsinh)
        # 1 when network outputs 0 (shift)
        total_width = keras.ops.arcsinh(keras.ops.softplus(parameters["total_width"] + self._shift))
        total_width = (self.default_width - self.min_width) * total_width + self.min_width

        total_height = keras.ops.arcsinh(keras.ops.softplus(parameters["total_height"] + self._shift))
        total_height = (self.default_height - self.min_height) * total_height + self.min_height

        bin_widths = keras.ops.softmax(parameters["bin_widths"], axis=-1)
        bin_widths = (total_width - self.bins * self.min_bin_width) * bin_widths + self.min_bin_width
        bin_heights = keras.ops.softmax(parameters["bin_heights"], axis=-1)
        bin_heights = (total_height - self.bins * self.min_bin_height) * bin_heights + self.min_bin_height

        # dy / dx
        affine_scale = total_height / total_width
        # y = a * x + b -> b = y - a * x
        affine_shift = bottom_edge - affine_scale * left_edge

        horizontal_edges = keras.ops.cumsum(bin_widths, axis=-1)
        horizontal_edges = pad(horizontal_edges, 0.0, 1, axis=-1, side="left")
        horizontal_edges = left_edge + horizontal_edges

        vertical_edges = keras.ops.cumsum(bin_heights, axis=-1)
        vertical_edges = pad(vertical_edges, 0.0, 1, axis=-1, side="left")
        vertical_edges = bottom_edge + vertical_edges

        derivatives = shifted_softplus(parameters["derivatives"])
        derivatives = pad(derivatives, affine_scale, 1, axis=-1, side="both")

        constrained_parameters = {
            "horizontal_edges": horizontal_edges,
            "vertical_edges": vertical_edges,
            "derivatives": derivatives,
            "affine_scale": keras.ops.squeeze(affine_scale, axis=-1),
            "affine_shift": keras.ops.squeeze(affine_shift, axis=-1),
        }

        return constrained_parameters

    def _forward(self, x: Tensor, parameters: dict[str, Tensor]) -> (Tensor, Tensor):
        # avoid side effects for mutable args
        parameters = parameters.copy()

        # affine transform for outside
        scale = parameters.pop("affine_scale")
        shift = parameters.pop("affine_shift")
        affine = scale * x + shift
        affine_log_jac = keras.ops.broadcast_to(keras.ops.log(scale), keras.ops.shape(affine))

        # spline transform for inside
        bins = searchsorted(parameters["horizontal_edges"], keras.ops.expand_dims(x, axis=-1))
        bins = keras.ops.squeeze(bins, axis=-1)
        inside = (bins > 0) & (bins <= self.bins)

        upper = bins
        lower = upper - 1

        # we need to mask out invalid bins to be backend-agnostic
        # this does not matter since we will overwrite these values with the affine values anyway
        upper = keras.ops.where(inside, upper, keras.ops.ones_like(upper))
        lower = keras.ops.where(inside, lower, keras.ops.zeros_like(lower))

        # need to expand the dimensions to match the shape of the parameters for take_along_axis
        upper = keras.ops.expand_dims(upper, axis=-1)
        lower = keras.ops.expand_dims(lower, axis=-1)

        edges = {
            "left": keras.ops.take_along_axis(parameters["horizontal_edges"], lower, axis=-1),
            "right": keras.ops.take_along_axis(parameters["horizontal_edges"], upper, axis=-1),
            "bottom": keras.ops.take_along_axis(parameters["vertical_edges"], lower, axis=-1),
            "top": keras.ops.take_along_axis(parameters["vertical_edges"], upper, axis=-1),
        }
        edges = {key: keras.ops.squeeze(value, axis=-1) for key, value in edges.items()}

        derivatives = {
            "left": keras.ops.take_along_axis(parameters["derivatives"], lower, axis=-1),
            "right": keras.ops.take_along_axis(parameters["derivatives"], upper, axis=-1),
        }
        derivatives = {key: keras.ops.squeeze(value, axis=-1) for key, value in derivatives.items()}

        parameters = {"edges": edges, "derivatives": derivatives}

        # compute the spline and jacobian
        spline, spline_log_jac = self.method_fn(x, **parameters)

        z = keras.ops.where(inside, spline, affine)
        log_jac = keras.ops.where(inside, spline_log_jac, affine_log_jac)

        log_det = keras.ops.sum(log_jac, axis=-1)

        return z, log_det

    def _inverse(self, z: Tensor, parameters: dict[str, Tensor]) -> (Tensor, Tensor):
        # avoid side effects for mutable args
        parameters = parameters.copy()

        # affine transform for outside
        scale = parameters.pop("affine_scale")
        shift = parameters.pop("affine_shift")
        affine = (z - shift) / scale
        affine_log_jac = keras.ops.broadcast_to(-keras.ops.log(scale), keras.ops.shape(affine))

        # spline transform for inside
        bins = searchsorted(parameters["vertical_edges"], keras.ops.expand_dims(z, axis=-1))
        bins = keras.ops.squeeze(bins, axis=-1)
        inside = (bins > 0) & (bins <= self.bins)

        upper = bins
        lower = upper - 1

        # we need to mask out invalid bins to be backend-agnostic
        # this does not matter since we will overwrite these values with the affine values anyway
        upper = keras.ops.where(inside, upper, keras.ops.ones_like(upper))
        lower = keras.ops.where(inside, lower, keras.ops.zeros_like(lower))

        # need to expand the dimensions to match the shape of the parameters for take_along_axis
        upper = keras.ops.expand_dims(upper, axis=-1)
        lower = keras.ops.expand_dims(lower, axis=-1)

        edges = {
            "left": keras.ops.take_along_axis(parameters["horizontal_edges"], lower, axis=-1),
            "right": keras.ops.take_along_axis(parameters["horizontal_edges"], upper, axis=-1),
            "bottom": keras.ops.take_along_axis(parameters["vertical_edges"], lower, axis=-1),
            "top": keras.ops.take_along_axis(parameters["vertical_edges"], upper, axis=-1),
        }
        edges = {key: keras.ops.squeeze(value, axis=-1) for key, value in edges.items()}

        derivatives = {
            "left": keras.ops.take_along_axis(parameters["derivatives"], lower, axis=-1),
            "right": keras.ops.take_along_axis(parameters["derivatives"], upper, axis=-1),
        }
        derivatives = {key: keras.ops.squeeze(value, axis=-1) for key, value in derivatives.items()}

        parameters = {"edges": edges, "derivatives": derivatives}

        # compute the spline and jacobian
        spline, spline_log_jac = self.method_fn(z, **parameters, inverse=True)

        x = keras.ops.where(inside, spline, affine)
        log_jac = keras.ops.where(inside, spline_log_jac, affine_log_jac)

        log_det = keras.ops.sum(log_jac, axis=-1)

        return x, log_det



from collections.abc import Sequence
import numpy as np

from bayesflow.types import Shape
from bayesflow.utils.decorators import allow_batch_size

from .simulator import Simulator


class SequentialSimulator(Simulator):
    """Combines multiple simulators into one, sequentially."""

    def __init__(self, simulators: Sequence[Simulator], expand_outputs: bool = True, replace_inputs: bool = True):
        """
        Initialize a SequentialSimulator.

        Parameters
        ----------
        simulators : Sequence[Simulator]
            A sequence of simulator instances to be executed sequentially. Each simulator should
            return dictionary outputs and may depend on outputs from previous simulators.
        expand_outputs : bool, optional
            If True, 1D output arrays are expanded with an additional dimension at the end.
            Default is True.
        replace_inputs : bool, optional
            If True, **kwargs are auto-batched and replace simulator outputs.
        """

        self.simulators = simulators
        self.expand_outputs = expand_outputs
        self.replace_inputs = replace_inputs

    @allow_batch_size
    def sample(self, batch_shape: Shape, **kwargs) -> dict[str, np.ndarray]:
        """
        Sample sequentially from the internal simulator.

        Parameters
        ----------
        batch_shape : Shape
            The shape of the batch to sample. Typically, a tuple indicating the number of samples,
            but it also accepts an int.
        **kwargs
            Additional keyword arguments passed to each simulator. These may include previously
            sampled outputs used as inputs for subsequent simulators.

        Returns
        -------
        data : dict of str to np.ndarray
            A dictionary containing the combined outputs from all simulators. Keys are output names
            and values are sampled arrays. If `expand_outputs` is True, 1D arrays are expanded to
            have shape (..., 1).
        """

        data = {}
        for simulator in self.simulators:
            data |= simulator.sample(batch_shape, **(kwargs | data))

            if self.replace_inputs:
                common_keys = set(data.keys()) & set(kwargs.keys())
                for key in common_keys:
                    value = kwargs.pop(key)
                    if isinstance(data[key], np.ndarray):
                        value = np.broadcast_to(value, data[key].shape)
                    data[key] = value

        if self.expand_outputs:
            data = {
                key: np.expand_dims(value, axis=-1) if np.ndim(value) == 1 else value for key, value in data.items()
            }

        return data



import numpy as np

from bayesflow.utils.serialization import serializable, serialize

from .elementwise_transform import ElementwiseTransform


@serializable
class Shift(ElementwiseTransform):
    def __init__(self, shift: np.typing.ArrayLike):
        self.shift = np.array(shift)

    def get_config(self) -> dict:
        return serialize({"shift": self.shift})

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return data + self.shift

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return data - self.shift



from collections.abc import Sequence, Mapping, Callable

import numpy as np

from ...utils.dict_utils import dicts_to_arrays


def posterior_contraction(
    estimates: Mapping[str, np.ndarray] | np.ndarray,
    targets: Mapping[str, np.ndarray] | np.ndarray,
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    aggregation: Callable = np.median,
) -> dict[str, any]:
    """
    Computes the posterior contraction (PC) from prior to posterior for the given samples.

    Parameters
    ----------
    estimates   : np.ndarray of shape (num_datasets, num_draws_post, num_variables)
        Posterior samples, comprising `num_draws_post` random draws from the posterior distribution
        for each data set from `num_datasets`.
    targets  : np.ndarray of shape (num_datasets, num_variables)
        Prior samples, comprising `num_datasets` ground truths.
    variable_keys : Sequence[str], optional (default = None)
       Select keys from the dictionaries provided in estimates and targets.
       By default, select all keys.
    variable_names : Sequence[str], optional (default = None)
        Optional variable names to show in the output.
    aggregation    : callable, optional (default = np.median)
        Function to aggregate the PC across draws. Typically `np.mean` or `np.median`.

    Returns
    -------
    result : dict
        Dictionary containing:

        - "values" : float or np.ndarray
            The aggregated posterior contraction per variable
        - "metric_name" : str
            The name of the metric ("Posterior Contraction").
        - "variable_names" : str
            The (inferred) variable names.

    Notes
    -----
    Posterior contraction measures the reduction in uncertainty from the prior to the posterior.
    Values close to 1 indicate strong contraction (high reduction in uncertainty), while values close to 0
    indicate low contraction.
    """

    samples = dicts_to_arrays(
        estimates=estimates,
        targets=targets,
        variable_keys=variable_keys,
        variable_names=variable_names,
    )

    post_vars = samples["estimates"].var(axis=1, ddof=1)
    prior_vars = samples["targets"].var(axis=0, keepdims=True, ddof=1)
    contraction = np.clip(1 - (post_vars / prior_vars), 0, 1)
    contraction = aggregation(contraction, axis=0)
    variable_names = samples["estimates"].variable_names
    return {"values": contraction, "metric_name": "Posterior Contraction", "variable_names": variable_names}



"""
A collection of :py:class:`~bayesflow.adapters.Adapter` transforms, which tell BayesFlow how to interpret your
:py:class:`~bayesflow.simulators.Simulator` output and plug it into neural networks for training and inference.
"""

from . import transforms
from .adapter import Adapter

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=["transforms"])



import keras

from bayesflow.types import Shape, Tensor
from bayesflow.utils.serialization import serializable

from ..invertible_layer import InvertibleLayer


@serializable
class FixedPermutation(InvertibleLayer):
    """
    Interface class for permutations with no learnable parameters. Child classes should
    create forward and inverse indices in the associated build() method.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.forward_indices = None
        self.inverse_indices = None

    def call(self, xz: Tensor, inverse: bool = False, **kwargs):
        if inverse:
            return self._inverse(xz)
        return self._forward(xz)

    def build(self, xz_shape: Shape, **kwargs) -> None:
        raise NotImplementedError

    def _forward(self, x: Tensor) -> (Tensor, Tensor):
        z = keras.ops.take(x, self.forward_indices, axis=-1)
        log_det = keras.ops.zeros(keras.ops.shape(x)[:-1])
        return z, log_det

    def _inverse(self, z: Tensor) -> (Tensor, Tensor):
        x = keras.ops.take(z, self.inverse_indices, axis=-1)
        log_det = keras.ops.zeros(keras.ops.shape(x)[:-1])
        return x, log_det



"""
This module provides :py:class:`~bayesflow.simulators.Simulator`, :py:func:`~bayesflow.simulators.make_simulator`,
and several other kinds of :py:class:`~bayesflow.simulators.Simulator` implementations, which serve as generative
mathematical models, or data generating processes, with their primary function being to sample data.
"""

from .sequential_simulator import SequentialSimulator
from .hierarchical_simulator import HierarchicalSimulator
from .lambda_simulator import LambdaSimulator
from .make_simulator import make_simulator
from .model_comparison_simulator import ModelComparisonSimulator
from .simulator import Simulator

from .benchmark_simulators import (
    BernoulliGLM,
    BernoulliGLMRaw,
    GaussianLinear,
    GaussianLinearUniform,
    GaussianMixture,
    InverseKinematics,
    LotkaVolterra,
    SIR,
    SLCP,
    SLCPDistractors,
    TwoMoons,
)

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])



class empty:
    """
    Placeholder value for arguments left empty

    Usage:

    def f(x=empty):
        if x is empty:
            # we know the user did not pass x
        if x is None:
            # the user could have passed None explicitly

    """



from keras.saving import register_keras_serializable as serializable

from bayesflow.types import Tensor
from bayesflow.utils import weighted_mean

from .scoring_rule import ScoringRule


@serializable(package="bayesflow.scores")
class ParametricDistributionScore(ScoringRule):
    r""":math:`S(\hat p_\phi, \theta; k) = -\log(\hat p_\phi(\theta))`

    Base class for scoring a predicted parametric probability distribution with the log-score
    of the probability of the realized value.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def log_prob(self, *args, **kwargs):
        raise NotImplementedError

    def sample(self, *args, **kwargs):
        raise NotImplementedError

    def score(self, estimates: dict[str, Tensor], targets: Tensor, weights: Tensor = None) -> Tensor:
        r"""
        Computes the log-score for a predicted parametric probability distribution given realized **targets**.

        :math:`S(\hat p_\phi, \theta; k) = -\log(\hat p_\phi(\theta))`
        """
        scores = -self.log_prob(x=targets, **estimates)
        score = weighted_mean(scores, weights)
        return score



from collections.abc import Sequence
import keras
import numpy as np

from bayesflow.types import Shape
from bayesflow.utils.decorators import allow_batch_size

from .simulator import Simulator


class HierarchicalSimulator(Simulator):
    def __init__(self, hierarchy: Sequence[Simulator]):
        """
        Initialize the hierarchical simulator with a sequence of simulators.

        Parameters
        ----------
        hierarchy : Sequence[Simulator]
            A sequence of simulator instances representing each level of the hierarchy.
            Each level's output is used as input for the next, with increasing batch dimensions.
        """
        self.hierarchy = hierarchy

    @allow_batch_size
    def sample(self, batch_shape: Shape, **kwargs) -> dict[str, np.ndarray]:
        """
        Sample from a hierarchy of simulators.

        Parameters
        ----------
        batch_shape : Shape
            A tuple where each element specifies the number of samples at the corresponding level
            of the hierarchy. The total batch size increases multiplicatively through the levels.
        **kwargs
            Additional keyword arguments passed to each simulator. These are combined with outputs
            from previous levels and repeated appropriately.

        Returns
        -------
        output_data : dict of str to np.ndarray
            A dictionary containing the outputs from the entire hierarchy. Outputs are reshaped to
            match the hierarchical batch shape, i.e., with shape equal to `batch_shape + original_shape`.
        """

        input_data = {}
        output_data = {}

        for level in range(len(self.hierarchy)):
            # repeat input data for the next level
            def repeat_level(x):
                return np.repeat(x, batch_shape[level], axis=0)

            input_data = keras.tree.map_structure(repeat_level, input_data)

            # query the simulator flat at the current level
            simulator = self.hierarchy[level]
            query_shape = (np.prod(batch_shape[: level + 1]),)
            data = simulator.sample(query_shape, **(kwargs | input_data))

            # input data needs to have a flat batch shape
            input_data |= data

            # output data needs the restored batch shape
            def restore_batch_shape(x):
                return np.reshape(x, batch_shape[: level + 1] + x.shape[1:])

            data = keras.tree.map_structure(restore_batch_shape, data)
            output_data |= data

        return output_data



from bayesflow.types import Tensor
from ..invertible_layer import InvertibleLayer


class Transform(InvertibleLayer):
    @property
    def params_per_dim(self) -> int:
        raise NotImplementedError

    def split_parameters(self, parameters: Tensor) -> dict[str, Tensor]:
        raise NotImplementedError

    def constrain_parameters(self, parameters: dict[str, Tensor]) -> dict[str, Tensor]:
        raise NotImplementedError

    def call(self, xz: Tensor, parameters: dict[str, Tensor], inverse: bool = False) -> (Tensor, Tensor):
        if inverse:
            return self._inverse(xz, parameters)
        return self._forward(xz, parameters)

    def _forward(self, x: Tensor, parameters: dict[str, Tensor]) -> (Tensor, Tensor):
        raise NotImplementedError

    def _inverse(self, z: Tensor, parameters: dict[str, Tensor]) -> (Tensor, Tensor):
        raise NotImplementedError



import keras
from keras import layers

from bayesflow.types import Tensor
from bayesflow.utils import check_lengths_same
from bayesflow.utils.serialization import serializable

from ..summary_network import SummaryNetwork

from .mab import MultiHeadAttentionBlock


@serializable
class FusionTransformer(SummaryNetwork):
    """Implements a more flexible version of the TimeSeriesTransformer that applies a series of self-attention layers
    followed by cross-attention between the representation and a learnable template summarized via a recurrent net."""

    def __init__(
        self,
        summary_dim: int = 16,
        embed_dims: tuple = (64, 64),
        num_heads: tuple = (4, 4),
        mlp_depths: tuple = (2, 2),
        mlp_widths: tuple = (128, 128),
        dropout: float = 0.05,
        mlp_activation: str = "gelu",
        kernel_initializer: str = "lecun_normal",
        use_bias: bool = True,
        layer_norm: bool = True,
        template_type: str = "lstm",
        bidirectional: bool = True,
        template_dim: int = 128,
        **kwargs,
    ):
        """Creates a fusion transformer used to flexibly compress time series and learn additional time embeddings
        using a recurrent neural network. If the time intervals vary across batches, it is highly recommended that
        your simulator also returns a "time" vector appended to the simulator outputs.

        Important: This network needs at least 2 transformer blocks and will generally be slower than the
        corresponding TimeSeriesTransformer.

        Parameters
        ----------
        summary_dim : int, optional (default - 16)
            Dimensionality of the final summary output.
        embed_dims  : tuple of int, optional (default - (64, 64))
            Dimensions of the keys, values, and queries for each attention block.
        num_heads   : tuple of int, optional (default - (4, 4))
            Number of attention heads for each embedding dimension.
        mlp_depths  : tuple of int, optional (default - (2, 2))
            Depth of the multi-layer perceptron (MLP) blocks for each component.
        mlp_widths  : tuple of int, optional (default - (128, 128))
            Width of each MLP layer in each block for each component.
        dropout     : float, optional (default - 0.05)
            Dropout rate applied to the attention and MLP layers. If set to None, no dropout is applied.
        mlp_activation : str, optional (default - 'gelu')
            Activation function used in the dense layers. Common choices include "relu", "elu", and "gelu".
        kernel_initializer : str, optional (default - 'lecun_normal')
            Initializer for the kernel weights matrix. Common choices include "glorot_uniform", "he_normal", etc.
        use_bias : bool, optional (default - True)
            Whether to include a bias term in the dense layers.
        layer_norm : bool, optional (default - True)
            Whether to apply layer normalization after the attention and MLP layers.
        t2v_embed_dim : int, optional (default - 8)
            The dimensionality of the Time2Vec embedding.
        template_type        : str or callable, optional, default: 'lstm'
            The many-to-one (learnable) transformation of the time series.
            if ``lstm``, an LSTM network will be used.
            if ``gru``, a GRU unit will be used.
        bidirectional        : bool, optional (default - False)
            Indicates whether the involved recurrent template network is bidirectional (i.e., forward
            and backward in time) or unidirectional (forward in time). Defaults to False, but may
            increase performance in some applications.
        template_dim         : int, optional (default - 128)
            Only used if ``template_type`` in ['lstm', 'gru']. The number of hidden
            units (equiv. output dimensions) of the recurrent network.
        time_axis     : int, optional (default - None)
            The time axis (e.g., -1 for last axis) from which to grab the time vector that goes into t2v.
        **kwargs : dict
            Additional keyword arguments passed to the base layer.
        """

        super().__init__(**kwargs)

        # Ensure all tuple-settings have the same length
        check_lengths_same(embed_dims, num_heads, mlp_depths, mlp_widths)

        # Construct a series of set-attention blocks
        self.attention_blocks = []
        for i in range(len(embed_dims)):
            layer_attention_settings = dict(
                dropout=dropout,
                mlp_activation=mlp_activation,
                kernel_initializer=kernel_initializer,
                use_bias=use_bias,
                layer_norm=layer_norm,
                num_heads=num_heads[i],
                embed_dim=embed_dims[i],
                mlp_depth=mlp_depths[i],
                mlp_width=mlp_widths[i],
            )

            block = MultiHeadAttentionBlock(**layer_attention_settings)
            self.attention_blocks.append(block)

        # A recurrent network will learn a dynamic many-to-one template
        if template_type.upper() == "LSTM":
            self.template_net = (
                layers.Bidirectional(layers.LSTM(template_dim // 2, dropout=dropout))
                if bidirectional
                else (layers.LSTM(template_dim, dropout=dropout))
            )
        elif template_type.upper() == "GRU":
            self.template_net = (
                layers.Bidirectional(layers.GRU(template_dim // 2, dropout=dropout))
                if bidirectional
                else (layers.GRU(template_dim, dropout=dropout))
            )
        else:
            raise ValueError("Argument `template_dim` should be in ['lstm', 'gru']")

        self.output_projector = keras.layers.Dense(summary_dim)
        self.summary_dim = summary_dim

    def call(self, input_sequence: Tensor, training: bool = False, **kwargs) -> Tensor:
        """Compresses the input sequence into a summary vector of size `summary_dim`.

        Parameters
        ----------
        input_sequence  : Tensor
            Input of shape (batch_size, sequence_length, input_dim)
        training        : boolean, optional (default - False)
            Passed to the optional internal dropout and spectral normalization
            layers to distinguish between train and test time behavior.
        **kwargs        : dict, optional (default - {})
            Additional keyword arguments passed to the internal attention layer,
            such as ``attention_mask`` or ``return_attention_scores``

        Returns
        -------
        out : Tensor
            Output of shape (batch_size, set_size, output_dim)
        """

        template = self.template_net(input_sequence, training=training)

        rep = input_sequence
        for layer in self.attention_blocks[:-1]:
            rep = layer(rep, rep, training=training, **kwargs)

        summary = self.attention_blocks[-1](keras.ops.expand_dims(template, axis=1), rep, training=training, **kwargs)
        summary = self.output_projector(keras.ops.squeeze(summary, axis=1))
        return summary



from collections.abc import Sequence

import keras

from bayesflow.types import Tensor
from bayesflow.utils import layer_kwargs
from bayesflow.utils import find_pooling
from bayesflow.utils.decorators import sanitize_input_shape
from bayesflow.utils.serialization import serializable

from ..mlp import MLP


@serializable
class InvariantLayer(keras.Layer):
    """Implements an invariant module performing a permutation-invariant transform.

    For details and rationale, see:

    [1] Bloem-Reddy, B., & Teh, Y. W. (2020). Probabilistic Symmetries and Invariant Neural Networks.
    J. Mach. Learn. Res., 21, 90-1. https://www.jmlr.org/papers/volume21/19-322/19-322.pdf
    """

    def __init__(
        self,
        mlp_widths_inner: Sequence[int] = (128, 128),
        mlp_widths_outer: Sequence[int] = (128, 128),
        activation: str = "gelu",
        kernel_initializer: str = "he_normal",
        dropout: int | float | None = 0.05,
        pooling: str = "mean",
        pooling_kwargs: dict = None,
        spectral_normalization: bool = False,
        **kwargs,
    ):
        """
        Initializes an invariant module representing a learnable permutation-invariant function with an option for
        learnable pooling.

        This module applies a two-stage transformation: an inner fully connected network processes individual
        set elements, followed by a pooling operation that aggregates features across the set. The pooled features are
        then passed through an outer fully connected network to produce the final invariant representation.

        The model supports different activation functions, dropout, and optional spectral normalization for stability.
        The pooling mechanism can be customized with additional arguments.

        Parameters
        ----------
        mlp_widths_inner : Sequence[int], optional
            Widths of the MLP layers applied before pooling. Default is (128, 128).
        mlp_widths_outer : Sequence[int], optional
            Widths of the MLP layers applied after pooling. Default is (128, 128).
        activation : str, optional
            Activation function applied in the MLP layers, such as "gelu". Default is "gelu".
        kernel_initializer : str, optional
            Initialization strategy for kernel weights, such as "he_normal". Default is "he_normal".
        dropout : int, float, or None, optional
            Dropout rate applied in the outer MLP layers. Default is 0.05.
        pooling : str, optional
            Type of pooling operation applied across set elements, such as "mean". Default is "mean".
        pooling_kwargs : dict, optional
            Additional keyword arguments for the pooling layer. Default is None.
        spectral_normalization : bool, optional
            Whether to apply spectral normalization to stabilize training. Default is False.
        """

        super().__init__(**layer_kwargs(kwargs))

        # Inner fully connected net for sum decomposition: inner( pooling( inner(set) ) )
        self.inner_fc = MLP(
            mlp_widths_inner,
            dropout=dropout,
            activation=activation,
            kernel_initializer=kernel_initializer,
            spectral_normalization=spectral_normalization,
        )

        self.outer_fc = MLP(
            mlp_widths_outer,
            dropout=dropout,
            activation=activation,
            kernel_initializer=kernel_initializer,
            spectral_normalization=spectral_normalization,
        )

        # Pooling function as keras layer for sum decomposition: inner( pooling( inner(set) ) )
        if pooling_kwargs is None:
            pooling_kwargs = {}

        self.pooling_layer = find_pooling(pooling, **pooling_kwargs)

    def call(self, input_set: Tensor, training: bool = False, **kwargs) -> Tensor:
        """Performs the forward pass of a learnable invariant transform.

        Parameters
        ----------
        input_set : Tensor
            Input of shape (batch_size,..., input_dim)
        training  : bool, optional, default - False
            Dictates the behavior of the optional dropout layers

        Returns
        -------
        set_summary : tf.Tensor
            Output of shape (batch_size,..., out_dim)
        """

        set_summary = self.inner_fc(input_set, training=training)
        set_summary = self.pooling_layer(set_summary, training=training)
        set_summary = self.outer_fc(set_summary, training=training)
        return set_summary

    @sanitize_input_shape
    def build(self, input_shape):
        self.call(keras.ops.zeros(input_shape))



from .bernoulli_glm import BernoulliGLM
from .bernoulli_glm_raw import BernoulliGLMRaw
from .gaussian_linear import GaussianLinear
from .gaussian_linear_uniform import GaussianLinearUniform
from .gaussian_mixture import GaussianMixture
from .inverse_kinematics import InverseKinematics
from .lotka_volterra import LotkaVolterra
from .sir import SIR
from .slcp import SLCP
from .slcp_distractors import SLCPDistractors
from .two_moons import TwoMoons



from typing import TypedDict

import keras

from bayesflow.types import Tensor


class Edges(TypedDict):
    left: Tensor
    right: Tensor
    bottom: Tensor
    top: Tensor


class Derivatives(TypedDict):
    left: Tensor
    right: Tensor


def _rational_quadratic_spline(
    x: Tensor, edges: Edges, derivatives: Derivatives, inverse: bool = False
) -> (Tensor, Tensor):
    # rename variables to match the paper:

    # $x^{(k)}$
    xk = edges["left"]

    # $x^{(k+1)}$
    xkp = edges["right"]

    # $y^{(k)}$
    yk = edges["bottom"]

    # $y^{(k+1)}$
    ykp = edges["top"]

    # $delta^{(k)}$
    dk = derivatives["left"]

    # $delta^{(k+1)}$
    dkp = derivatives["right"]

    # commonly used values
    dx = xkp - xk
    dy = ykp - yk
    sk = dy / dx

    if not inverse:
        xi = (x - xk) / dx

        # Eq. 4 in the paper
        numerator = dy * (sk * xi**2 + dk * xi * (1 - xi))
        denominator = sk + (dkp + dk - 2 * sk) * xi * (1 - xi)
        result = yk + numerator / denominator
    else:
        # rename for clarity
        y = x

        # Eq. 6-8 in the paper
        a = dy * (sk - dk) + (y - yk) * (dkp + dk - 2 * sk)
        b = dy * dk - (y - yk) * (dkp + dk - 2 * sk)
        c = -sk * (y - yk)

        # Eq. 29 in the appendix of the paper
        discriminant = b**2 - 4 * a * c

        # the discriminant must be positive, even when the spline is called out of bounds
        discriminant = keras.ops.maximum(discriminant, 0)

        xi = 2 * c / (-b - keras.ops.sqrt(discriminant))
        result = xi * dx + xk

    # Eq 5 in the paper
    numerator = sk**2 * (dkp * xi**2 + 2 * sk * xi * (1 - xi) + dk * (1 - xi) ** 2)
    denominator = (sk + (dkp + dk - 2 * sk) * xi * (1 - xi)) ** 2
    log_jac = keras.ops.log(numerator) - keras.ops.log(denominator)

    if inverse:
        log_jac = -log_jac

    return result, log_jac



import keras

from bayesflow.types import Tensor
from bayesflow.utils import check_lengths_same
from bayesflow.utils.serialization import serializable

from ..summary_network import SummaryNetwork

from .sab import SetAttentionBlock
from .isab import InducedSetAttentionBlock
from .pma import PoolingByMultiHeadAttention


@serializable
class SetTransformer(SummaryNetwork):
    """Implements the set transformer architecture from [1] which ultimately represents
    a learnable permutation-invariant function. Designed to naturally model interactions in
    the input set, which may be hard to capture with the simpler ``DeepSet`` architecture.

    [1] Lee, J., Lee, Y., Kim, J., Kosiorek, A., Choi, S., & Teh, Y. W. (2019).
        Set transformer: A framework for attention-based permutation-invariant neural networks.
        In International conference on machine learning (pp. 3744-3753). PMLR.

    Note: Currently works only on 3D inputs but can easily be expanded by using ``keras.layers.TimeDistributed``.
    """

    def __init__(
        self,
        summary_dim: int = 16,
        embed_dims: tuple = (64, 64),
        num_heads: tuple = (4, 4),
        mlp_depths: tuple = (2, 2),
        mlp_widths: tuple = (128, 128),
        num_seeds: int = 1,
        dropout: float = 0.05,
        mlp_activation: str = "gelu",
        kernel_initializer: str = "lecun_normal",
        use_bias: bool = True,
        layer_norm: bool = True,
        num_inducing_points: int = None,
        seed_dim: int = None,
        **kwargs,
    ):
        """
        Creates a many-to-one permutation-invariant encoder, typically used as a summary net for embedding set-based,
        (i.e., exchangeable or IID) data. Use a TimeSeriesTransformer or a FusionTransformer for non-IID data.

        The number of multi-head attention block is inferred from the length of `embed_dims` tuple.

        Parameters
        ----------
        summary_dim : int, optional (default - 16)
            Dimensionality of the final summary output.
        embed_dims  : tuple of int, optional (default - (64, 64))
            Dimensions of the keys, values, and queries for each attention block.
        num_heads   : tuple of int, optional (default - (4, 4))
            Number of attention heads for each embedding dimension.
        mlp_depths  : tuple of int, optional (default - (2, 2))
            Depth of the multi-layer perceptron (MLP) blocks for each component.
        mlp_widths  : tuple of int, optional (default - (128, 128))
            Width of each MLP layer in each block for each component.
        num_seeds   : int, optional (default - 1)
            Number of seeds to use for embedding.
        dropout     : float, optional (default - 0.05)
            Dropout rate applied to the attention and MLP layers. If set to None, no dropout is applied.
        mlp_activation : str, optional (default - 'gelu')
            Activation function used in the dense layers. Common choices include "relu", "elu", and "gelu".
        kernel_initializer : str, optional (default - 'lecun_normal')
            Initializer for the kernel weights matrix. Common choices include "glorot_uniform", "he_normal", etc.
        use_bias : bool, optional (default - True)
            Whether to include a bias term in the dense layers.
        layer_norm : bool, optional (default - True)
            Whether to apply layer normalization after the attention and MLP layers.
        num_inducing_points : int or None, optional (default - None)
            Number of inducing points used, if applicable. If set to None, this option is disabled.
        seed_dim : int or None, optional (default - None)
            Dimensionality of the seed embeddings. If None, it defaults to `summary_dim`.
        **kwargs : dict
            Additional keyword arguments passed to the base layer.
        """

        super().__init__(**kwargs)

        check_lengths_same(embed_dims, num_heads, mlp_depths, mlp_widths)

        num_attention_layers = len(embed_dims)

        # Construct a series of set-attention blocks
        self.attention_blocks = keras.Sequential()

        global_attention_settings = dict(
            dropout=dropout,
            mlp_activation=mlp_activation,
            kernel_initializer=kernel_initializer,
            use_bias=use_bias,
            layer_norm=layer_norm,
        )

        for i in range(num_attention_layers):
            layer_attention_settings = dict(
                num_heads=num_heads[i],
                embed_dim=embed_dims[i],
                mlp_depth=mlp_depths[i],
                mlp_width=mlp_widths[i],
            )

            if num_inducing_points is None:
                block = SetAttentionBlock(**(global_attention_settings | layer_attention_settings))
            else:
                isab_settings = dict(num_inducing_points=num_inducing_points)
                block = InducedSetAttentionBlock(
                    **(global_attention_settings | layer_attention_settings | isab_settings)
                )

            self.attention_blocks.add(block)

        # Pooling will be applied as a final step to the abstract representations obtained from set attention
        pooling_settings = dict(
            num_heads=num_heads[-1],
            embed_dim=embed_dims[-1],
            mlp_depth=mlp_depths[-1],
            mlp_width=mlp_widths[-1],
            seed_dim=seed_dim,
            num_seeds=num_seeds,
        )
        self.pooling_by_attention = PoolingByMultiHeadAttention(**(global_attention_settings | pooling_settings))
        self.output_projector = keras.layers.Dense(summary_dim)

        self.summary_dim = summary_dim

    def call(self, input_set: Tensor, training: bool = False, **kwargs) -> Tensor:
        """Compresses the input sequence into a summary vector of size `summary_dim`.

        Parameters
        ----------
        input_set  : Tensor (e.g., np.ndarray, tf.Tensor, ...)
            Input of shape (batch_size, set_size, input_dim)
        training   : boolean, optional (default - False)
            Passed to the optional internal dropout and spectral normalization
            layers to distinguish between train and test time behavior.
        **kwargs   : dict, optional (default - {})
            Additional keyword arguments passed to the internal attention layer,
            such as ``attention_mask`` or ``return_attention_scores``

        Returns
        -------
        out : Tensor
            Output of shape (batch_size, set_size, output_dim)
        """
        summary = self.attention_blocks(input_set, training=training, **kwargs)
        summary = self.pooling_by_attention(summary, training=training, **kwargs)
        summary = self.output_projector(summary)
        return summary



import numpy as np

from bayesflow.utils.serialization import serializable, serialize

from .elementwise_transform import ElementwiseTransform


@serializable
class Scale(ElementwiseTransform):
    def __init__(self, scale: np.typing.ArrayLike):
        self.scale = np.array(scale)

    def get_config(self) -> dict:
        return serialize({"scale": self.scale})

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return data * self.scale

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return data / self.scale



import keras

from bayesflow.types import Shape
from bayesflow.utils.serialization import serializable

from .fixed_permutation import FixedPermutation


@serializable
class Swap(FixedPermutation):
    def build(self, xz_shape: Shape, **kwargs) -> None:
        shift = xz_shape[-1] // 2
        forward_indices = keras.ops.roll(keras.ops.arange(xz_shape[-1]), shift=shift)
        inverse_indices = keras.ops.argsort(forward_indices)

        self.forward_indices = self.add_variable(
            shape=(xz_shape[-1],),
            initializer=keras.initializers.Constant(forward_indices),
            trainable=False,
            dtype="int",
        )

        self.inverse_indices = self.add_variable(
            shape=(xz_shape[-1],),
            initializer=keras.initializers.Constant(inverse_indices),
            trainable=False,
            dtype="int",
        )



from collections.abc import Sequence

import numpy as np
from keras import ops

from ...utils.exceptions import ShapeError
from ...utils.classification import calibration_curve


def expected_calibration_error(
    estimates: np.ndarray,
    targets: np.ndarray,
    model_names: Sequence[str] = None,
    num_bins: int = 10,
    return_probs: bool = False,
) -> dict[str, any]:
    """
    Estimates the expected calibration error (ECE) of a model comparison network according to [1].

    [1] Naeini, M. P., Cooper, G., & Hauskrecht, M. (2015). Obtaining well calibrated probabilities using
    Bayesian binning. In Proceedings of the AAAI conference on artificial intelligence (Vol. 29, No. 1).

    Notes
    -----
    Make sure that ``targets`` are **one-hot encoded** classes (i.e., model indices)!

    Parameters
    ----------
    estimates      : array of shape (num_sim, num_models)
        The predicted posterior model probabilities.
    targets      : array of shape (num_sim, num_models)
        The one-hot-encoded true model indices.
    model_names : Sequence[str], optional (default = None)
        Optional model names to show in the output. By default, models are called "M_" + model index.
    num_bins    : int, optional, default: 10
        The number of bins to use for the calibration curves (and marginal histograms).
        Passed into ``bayesflow.utils.calibration_curve()``.
    return_probs : bool (default = False)
        Do you want to obtain the output of ``bayesflow.utils.calibration_curve()``?

    Returns
    -------
    result : dict
        Dictionary containing:
        - "values" : np.ndarray
            The expected calibration error per model
        - "metric_name" : str
            The name of the metric ("Expected Calibration Error").
        - "model_names" : str
            The (inferred) variable names.
        - "probs_true": (optional) list[np.ndarray]:
            Outputs of ``bayesflow.utils.calibration.calibration_curve()`` per model
        - "probs_pred": (optional) list[np.ndarray]:
            Outputs of ``bayesflow.utils.calibration.calibration_curve()`` per model
    """

    # Convert tensors to numpy, if passed
    estimates = ops.convert_to_numpy(estimates)
    targets = ops.convert_to_numpy(targets)

    if estimates.shape != targets.shape:
        raise ShapeError("`estimates` and `targets` must have the same shape.")

    if model_names is None:
        model_names = ["M_" + str(i) for i in range(estimates.shape[-1])]
    elif len(model_names) != estimates.shape[-1]:
        raise ShapeError("There must be exactly one `model_name` for each model in `estimates`")

    # Extract number of models and prepare containers
    ece = []
    probs_true = []
    probs_pred = []

    targets = targets.argmax(axis=-1)

    # Loop for each model and compute calibration errs per bin
    for model_index in range(estimates.shape[-1]):
        y_true = (targets == model_index).astype(np.float32)
        y_prob = estimates[..., model_index]
        prob_true, prob_pred = calibration_curve(y_true, y_prob, num_bins=num_bins)

        # Compute ECE by weighting bin errors by bin size
        bins = np.linspace(0.0, 1.0, num_bins + 1)
        binids = np.searchsorted(bins[1:-1], y_prob)
        bin_total = np.bincount(binids, minlength=len(bins))
        nonzero = bin_total != 0
        error = np.sum(np.abs(prob_true - prob_pred) * (bin_total[nonzero] / len(y_true)))

        ece.append(error)
        probs_true.append(prob_true)
        probs_pred.append(prob_pred)

    output = dict(values=np.array(ece), metric_name="Expected Calibration Error", model_names=model_names)

    if return_probs:
        output["probs_true"] = probs_true
        output["probs_pred"] = probs_pred

    return output



import keras
import pytest

from bayesflow.utils import optimal_transport
from tests.utils import assert_allclose


@pytest.mark.jax
def test_jit_compile():
    import jax

    x = keras.random.normal((128, 8), seed=0)
    y = keras.random.normal((128, 8), seed=1)

    ot = jax.jit(optimal_transport, static_argnames=["regularization", "seed"])
    ot(x, y, regularization=1.0, seed=0, max_steps=10)


@pytest.mark.parametrize("method", ["log_sinkhorn", "sinkhorn"])
def test_shapes(method):
    x = keras.random.normal((128, 8), seed=0)
    y = keras.random.normal((128, 8), seed=1)

    ox, oy = optimal_transport(x, y, regularization=1.0, seed=0, max_steps=10, method=method)

    assert keras.ops.shape(ox) == keras.ops.shape(x)
    assert keras.ops.shape(oy) == keras.ops.shape(y)


def test_transport_cost_improves():
    x = keras.random.normal((128, 2), seed=0)
    y = keras.random.normal((128, 2), seed=1)

    before_cost = keras.ops.sum(keras.ops.norm(x - y, axis=-1))

    x, y = optimal_transport(x, y, regularization=0.1, seed=0, max_steps=1000)

    after_cost = keras.ops.sum(keras.ops.norm(x - y, axis=-1))

    assert after_cost < before_cost


@pytest.mark.skip(reason="too unreliable")
def test_assignment_is_optimal():
    x = keras.random.normal((16, 2), seed=0)
    p = keras.random.shuffle(keras.ops.arange(keras.ops.shape(x)[0]), seed=0)
    optimal_assignments = keras.ops.argsort(p)

    y = x[p]

    x, y, assignments = optimal_transport(x, y, regularization=0.1, seed=0, max_steps=10_000, return_assignments=True)

    assert_allclose(assignments, optimal_assignments)


def test_assignment_aligns_with_pot():
    try:
        from ot.bregman import sinkhorn_log
    except (ImportError, ModuleNotFoundError):
        pytest.skip("Need to install POT to run this test.")

    x = keras.random.normal((16, 2), seed=0)
    p = keras.random.shuffle(keras.ops.arange(keras.ops.shape(x)[0]), seed=0)
    y = x[p]

    a = keras.ops.ones(keras.ops.shape(x)[0])
    b = keras.ops.ones(keras.ops.shape(y)[0])
    M = x[:, None] - y[None, :]
    M = keras.ops.norm(M, axis=-1)

    pot_plan = sinkhorn_log(a, b, M, reg=1e-3, numItermax=10_000, stopThr=1e-99)
    pot_assignments = keras.random.categorical(pot_plan, num_samples=1, seed=0)
    pot_assignments = keras.ops.squeeze(pot_assignments, axis=-1)

    _, _, assignments = optimal_transport(x, y, regularization=1e-3, seed=0, max_steps=10_000, return_assignments=True)

    assert_allclose(pot_assignments, assignments)



import keras
import numpy as np
import os
import pathlib as pl

from bayesflow.adapters import Adapter
from bayesflow.utils import tree_stack, pickle_load


class DiskDataset(keras.utils.PyDataset):
    """
    A dataset used to load pre-simulated files from disk.
    The training strategy will be offline.

    By default, the expected file structure is as follows:
    root
    ├── ...
    ├── sample_1.[ext]
    ├── ...
    └── sample_n.[ext]

    where each file contains a complete sample (e.g., a dictionary of numpy arrays) or
    is converted into a complete sample using a custom loader function.
    """

    def __init__(
        self,
        root: os.PathLike,
        *,
        pattern: str = "*.pkl",
        batch_size: int,
        load_fn: callable = None,
        adapter: Adapter | None,
        stage: str = "training",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.batch_size = batch_size
        self.root = pl.Path(root)
        self.load_fn = load_fn or pickle_load
        self.adapter = adapter
        self.files = list(map(str, self.root.glob(pattern)))
        self.stage = stage

        self.shuffle()

    def __getitem__(self, item) -> dict[str, np.ndarray]:
        if not 0 <= item < self.num_batches:
            raise IndexError(f"Index {item} is out of bounds for dataset with {self.num_batches} batches.")

        files = self.files[item * self.batch_size : (item + 1) * self.batch_size]

        batch = []
        for file in files:
            batch.append(self.load_fn(file))

        batch = tree_stack(batch)

        if self.adapter is not None:
            batch = self.adapter(batch, stage=self.stage)

        return batch

    def on_epoch_end(self):
        self.shuffle()

    @property
    def num_batches(self):
        return int(np.ceil(len(self.files) / self.batch_size))

    def shuffle(self):
        np.random.shuffle(self.files)



import numpy as np
import pytest


@pytest.fixture()
def batch_size():
    return 16


@pytest.fixture(params=[False, True], autouse=True)
def use_batched(request):
    return request.param


@pytest.fixture(params=[False, True], autouse=True)
def use_numpy(request):
    return request.param


@pytest.fixture(params=[False, True], autouse=True)
def use_squeezed(request):
    return request.param


@pytest.fixture()
def bernoulli_glm():
    from bayesflow.simulators import BernoulliGLM

    return BernoulliGLM()


@pytest.fixture()
def bernoulli_glm_raw():
    from bayesflow.simulators import BernoulliGLMRaw

    return BernoulliGLMRaw()


@pytest.fixture()
def gaussian_linear():
    from bayesflow.simulators import GaussianLinear

    return GaussianLinear()


@pytest.fixture()
def gaussian_linear_n_obs():
    from bayesflow.simulators import GaussianLinear

    return GaussianLinear(n_obs=5)


@pytest.fixture()
def gaussian_linear_uniform():
    from bayesflow.simulators import GaussianLinearUniform

    return GaussianLinearUniform()


@pytest.fixture()
def gaussian_linear_uniform_n_obs():
    from bayesflow.simulators import GaussianLinearUniform

    return GaussianLinearUniform(n_obs=5)


@pytest.fixture(
    params=["gaussian_linear", "gaussian_linear_n_obs", "gaussian_linear_uniform", "gaussian_linear_uniform_n_obs"]
)
def gaussian_linear_simulator(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def gaussian_mixture():
    from bayesflow.simulators import GaussianMixture

    return GaussianMixture()


@pytest.fixture()
def inverse_kinematics():
    from bayesflow.simulators import InverseKinematics

    return InverseKinematics()


@pytest.fixture()
def lotka_volterra():
    from bayesflow.simulators import LotkaVolterra

    return LotkaVolterra()


@pytest.fixture()
def sir():
    from bayesflow.simulators import SIR

    return SIR()


@pytest.fixture()
def slcp():
    from bayesflow.simulators import SLCP

    return SLCP()


@pytest.fixture()
def slcp_distractors():
    from bayesflow.simulators import SLCPDistractors

    return SLCPDistractors()


@pytest.fixture()
def composite_two_moons():
    from bayesflow.simulators import make_simulator

    def parameters():
        parameters = np.random.uniform(-1.0, 1.0, size=2)
        return dict(parameters=parameters)

    def observables(parameters):
        r = np.random.normal(0.1, 0.01)
        alpha = np.random.uniform(-0.5 * np.pi, 0.5 * np.pi)
        x1 = -np.abs(parameters[0] + parameters[1]) / np.sqrt(2.0) + r * np.cos(alpha) + 0.25
        x2 = (-parameters[0] + parameters[1]) / np.sqrt(2.0) + r * np.sin(alpha)
        return dict(observables=np.stack([x1, x2]))

    return make_simulator([parameters, observables])


@pytest.fixture()
def two_moons():
    from bayesflow.simulators import TwoMoons

    return TwoMoons()


@pytest.fixture(
    params=[
        "composite_two_moons",
        "two_moons",
    ]
)
def two_moons_simulator(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def composite_gaussian():
    from bayesflow.simulators import make_simulator

    def context():
        n = np.random.randint(10, 100)
        return dict(n=n)

    def prior():
        mu = np.random.normal(0, 1)
        return dict(mu=mu)

    def likelihood(mu, n):
        y = np.random.normal(mu, 1, n)
        return dict(y=y)

    return make_simulator([prior, likelihood], meta_fn=context)


@pytest.fixture()
def fixed_n():
    return 5


@pytest.fixture()
def fixed_mu():
    return 100


@pytest.fixture(
    params=[
        "bernoulli_glm",
        "bernoulli_glm_raw",
        "gaussian_linear",
        "gaussian_linear_n_obs",
        "gaussian_linear_uniform",
        "gaussian_linear_uniform_n_obs",
        "gaussian_mixture",
        "inverse_kinematics",
        "lotka_volterra",
        "sir",
        "slcp",
        "slcp_distractors",
        "composite_two_moons",
        "two_moons",
    ]
)
def simulator(request):
    return request.getfixturevalue(request.param)



from collections.abc import Mapping

import multiprocessing as mp

import keras

from bayesflow.adapters import Adapter
from bayesflow.datasets import OnlineDataset
from bayesflow.simulators import Simulator
from bayesflow.utils import find_batch_size, filter_kwargs, logging

from .backend_approximators import BackendApproximator


class Approximator(BackendApproximator):
    def build(self, data_shapes: any) -> None:
        mock_data = keras.tree.map_structure(keras.ops.zeros, data_shapes)
        self.build_from_data(mock_data)

    @classmethod
    def build_adapter(cls, **kwargs) -> Adapter:
        # implemented by each respective architecture
        raise NotImplementedError

    def build_from_data(self, data: Mapping[str, any]) -> None:
        self.compute_metrics(**filter_kwargs(data, self.compute_metrics), stage="training")
        self.built = True

    @classmethod
    def build_dataset(
        cls,
        *,
        batch_size: int = "auto",
        num_batches: int,
        adapter: Adapter = "auto",
        memory_budget: str | int = "auto",
        simulator: Simulator,
        workers: int = "auto",
        use_multiprocessing: bool = False,
        max_queue_size: int = 32,
        **kwargs,
    ) -> OnlineDataset:
        if batch_size == "auto":
            batch_size = find_batch_size(memory_budget=memory_budget, sample=simulator.sample((1,)))
            logging.info(f"Using a batch size of {batch_size}.")

        if adapter == "auto":
            adapter = cls.build_adapter(**filter_kwargs(kwargs, cls.build_adapter))

        if workers == "auto":
            workers = mp.cpu_count()
            logging.info(f"Using {workers} data loading workers.")

        workers = workers or 1

        return OnlineDataset(
            simulator=simulator,
            batch_size=batch_size,
            num_batches=num_batches,
            adapter=adapter,
            workers=workers,
            use_multiprocessing=use_multiprocessing,
            max_queue_size=max_queue_size,
        )

    def fit(self, *, dataset: keras.utils.PyDataset = None, simulator: Simulator = None, **kwargs):
        """
        Trains the approximator on the provided dataset or on-demand data generated from the given simulator.
        If `dataset` is not provided, a dataset is built from the `simulator`.
        If the model has not been built, it will be built using a batch from the dataset.

        Parameters
        ----------
        dataset : keras.utils.PyDataset, optional
            A dataset containing simulations for training. If provided, `simulator` must be None.
        simulator : Simulator, optional
            A simulator used to generate a dataset. If provided, `dataset` must be None.
        **kwargs
            Additional keyword arguments passed to `keras.Model.fit()`, including (see also `build_dataset`):

            batch_size : int or None, default='auto'
                Number of samples per gradient update. Do not specify if `dataset` is provided as a
                `keras.utils.PyDataset`, `tf.data.Dataset`, `torch.utils.data.DataLoader`, or a generator function.
            epochs : int, default=1
                Number of epochs to train the model.
            verbose : {"auto", 0, 1, 2}, default="auto"
                Verbosity mode. 0 = silent, 1 = progress bar, 2 = one line per epoch.
            callbacks : list of keras.callbacks.Callback, optional
                List of callbacks to apply during training.
            validation_split : float, optional
                Fraction of training data to use for validation (only supported if `dataset` consists of NumPy arrays
                or tensors).
            validation_data : tuple or dataset, optional
                Data for validation, overriding `validation_split`.
            shuffle : bool, default=True
                Whether to shuffle the training data before each epoch (ignored for dataset generators).
            initial_epoch : int, default=0
                Epoch at which to start training (useful for resuming training).
            steps_per_epoch : int or None, optional
                Number of steps (batches) before declaring an epoch finished.
            validation_steps : int or None, optional
                Number of validation steps per validation epoch.
            validation_batch_size : int or None, optional
                Number of samples per validation batch (defaults to `batch_size`).
            validation_freq : int, default=1
                Specifies how many training epochs to run before performing validation.

        Returns
        -------
        keras.callbacks.History
            A history object containing the training loss and metrics values.

        Raises
        ------
        ValueError
            If both `dataset` and `simulator` are provided or neither is provided.
        """

        if dataset is None:
            if simulator is None:
                raise ValueError("Received no data to fit on. Please provide either a dataset or a simulator.")

            logging.info(f"Building dataset from simulator instance of {simulator.__class__.__name__}.")
            dataset = self.build_dataset(simulator=simulator, **filter_kwargs(kwargs, self.build_dataset))
        else:
            if simulator is not None:
                raise ValueError(
                    "Received conflicting arguments. Please provide either a dataset or a simulator, but not both."
                )

            logging.info(f"Fitting on dataset instance of {dataset.__class__.__name__}.")

        if not self.built:
            logging.info("Building on a test batch.")
            mock_data = dataset[0]
            mock_data = keras.tree.map_structure(keras.ops.convert_to_tensor, mock_data)
            self.build_from_data(mock_data)

        return super().fit(dataset=dataset, **kwargs)



import keras

from bayesflow.utils import model_kwargs


class InvertibleLayer(keras.Model):
    def __init__(self, **kwargs):
        super().__init__(**model_kwargs(kwargs))

    def call(self, *args, **kwargs):
        # we cannot provide a default implementation for this
        #  because the signature of layer.call() is used to
        #  determine the arguments to layer.build()
        raise NotImplementedError

    def _forward(self, *args, **kwargs):
        raise NotImplementedError

    def _inverse(self, *args, **kwargs):
        raise NotImplementedError



from .affine_transform import AffineTransform
from .spline_transform import SplineTransform
from .transform import Transform


def find_transform(transform: str | Transform | type(Transform), **kwargs) -> Transform:
    if isinstance(transform, Transform):
        return transform
    if isinstance(transform, type):
        return transform()

    match transform.lower():
        case "affine":
            return AffineTransform()
        case "spline":
            return SplineTransform(**kwargs)
        case str() as unknown_transform:
            raise ValueError(f"Unknown transform: '{unknown_transform}'")
        case other:
            raise TypeError(f"Unknown transform type: {other}")



from collections.abc import Sequence, Mapping

import numpy as np
import pandas as pd
import seaborn as sns

import matplotlib.pyplot as plt

from bayesflow.utils import logging
from bayesflow.utils.dict_utils import dicts_to_arrays


def pairs_samples(
    samples: Mapping[str, np.ndarray] | np.ndarray = None,
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    height: float = 2.5,
    color: str | tuple = "#132a70",
    alpha: float = 0.9,
    label_fontsize: int = 14,
    tick_fontsize: int = 12,
    **kwargs,
) -> sns.PairGrid:
    """
    A more flexible pair plot function for multiple distributions based upon
    collected samples.

    Parameters
    ----------
    samples     : dict[str, Tensor], default: None
        Sample draws from any dataset
    variable_keys       : list or None, optional, default: None
       Select keys from the dictionary provided in samples.
       By default, select all keys.
    variable_names    : list or None, optional, default: None
        The parameter names for nice plot titles. Inferred if None
    height      : float, optional, default: 2.5
        The height of the pair plot
    color       : str, optional, default : '#8f2727'
        The color of the plot
    alpha       : float in [0, 1], optional, default: 0.9
        The opacity of the plot
    label_fontsize    : int, optional, default: 14
        The font size of the x and y-label texts (parameter names)
    tick_fontsize     : int, optional, default: 12
        The font size of the axis ticklabels
    **kwargs    : dict, optional
        Additional keyword arguments passed to the sns.PairGrid constructor
    """

    plot_data = dicts_to_arrays(
        estimates=samples,
        variable_keys=variable_keys,
        variable_names=variable_names,
    )

    g = _pairs_samples(
        plot_data=plot_data,
        height=height,
        color=color,
        alpha=alpha,
        label_fontsize=label_fontsize,
        tick_fontsize=tick_fontsize,
    )

    return g


def _pairs_samples(
    plot_data: dict,
    height: float = 2.5,
    color: str | tuple = "#132a70",
    color2: str | tuple = "gray",
    alpha: float = 0.9,
    label_fontsize: int = 14,
    tick_fontsize: int = 12,
    legend_fontsize: int = 14,
    **kwargs,
) -> sns.PairGrid:
    # internal version of pairs_samples creating the seaborn plot

    # Parameters
    # ----------
    # plot_data   : output of bayesflow.utils.dict_utils.dicts_to_arrays
    # other arguments are documented in pairs_samples

    estimates_shape = plot_data["estimates"].shape
    if len(estimates_shape) != 2:
        raise ValueError(
            f"Samples for a single distribution should be a matrix, but "
            f"your samples array has a shape of {estimates_shape}."
        )

    variable_names = plot_data["estimates"].variable_names

    # Convert samples to pd.DataFrame
    if plot_data["priors"] is not None:
        # differentiate posterior from prior draws
        # row bind posterior and prior draws
        samples = np.vstack((plot_data["priors"], plot_data["estimates"]))
        data_to_plot = pd.DataFrame(samples, columns=variable_names)

        # ensure that the source of the samples is stored
        source_prior = np.repeat("Prior", plot_data["priors"].shape[0])
        source_post = np.repeat("Posterior", plot_data["estimates"].shape[0])
        data_to_plot["_source"] = np.concatenate((source_prior, source_post))
        data_to_plot["_source"] = pd.Categorical(data_to_plot["_source"], categories=["Prior", "Posterior"])

        # initialize plot
        g = sns.PairGrid(
            data_to_plot,
            height=height,
            hue="_source",
            palette=[color2, color],
            **kwargs,
        )

        # ensures that color doesn't overwrite palette
        color = None

    else:
        # plot just the one set of distributions
        data_to_plot = pd.DataFrame(plot_data["estimates"], columns=variable_names)

        # initialize plot
        g = sns.PairGrid(data_to_plot, height=height, **kwargs)

    # add histograms + KDEs to the diagonal
    g.map_diag(
        histplot_twinx,
        fill=True,
        kde=True,
        color=color,
        alpha=alpha,
        stat="density",
        common_norm=False,
    )

    # add scatterplots to the upper diagonal
    g.map_upper(sns.scatterplot, alpha=0.6, s=40, edgecolor="k", color=color, lw=0)

    # add KDEs to the lower diagonal
    try:
        g.map_lower(sns.kdeplot, fill=True, color=color, alpha=alpha)
    except Exception as e:
        logging.exception("KDE failed due to the following exception:\n" + repr(e) + "\nSubstituting scatter plot.")
        g.map_lower(sns.scatterplot, alpha=0.6, s=40, edgecolor="k", color=color, lw=0)

    # need to add legend here such that colors are recognized
    if plot_data["priors"] is not None:
        g.add_legend(fontsize=legend_fontsize, loc="center right")
        g._legend.set_title(None)

    # Generate grids
    dim = g.axes.shape[0]
    for i in range(dim):
        for j in range(dim):
            g.axes[i, j].grid(alpha=0.5)
            g.axes[i, j].set_axisbelow(True)

    dim = g.axes.shape[0]
    for i in range(dim):
        # Modify tick sizes
        for j in range(i + 1):
            g.axes[i, j].tick_params(axis="both", which="major", labelsize=tick_fontsize)
            g.axes[i, j].tick_params(axis="both", which="minor", labelsize=tick_fontsize)

        # adjust font size of labels
        # the labels themselves remain the same as before, i.e., variable_names
        g.axes[i, 0].set_ylabel(variable_names[i], fontsize=label_fontsize)
        g.axes[dim - 1, i].set_xlabel(variable_names[i], fontsize=label_fontsize)

    # Return figure
    g.tight_layout()

    return g


# create a histogram plot on a twin y axis
# this ensures that the y scaling of the diagonal plots
# in independent of the y scaling of the off-diagonal plots
def histplot_twinx(x, **kwargs):
    # Create a twin axis
    ax2 = plt.gca().twinx()

    # create a histogram on the twin axis
    sns.histplot(x, **kwargs, ax=ax2)

    # make the twin axis invisible
    plt.gca().spines["right"].set_visible(False)
    plt.gca().spines["top"].set_visible(False)
    ax2.set_ylabel("")
    ax2.set_yticks([])
    ax2.set_yticklabels([])

    return None



import keras
import logging
from functools import lru_cache


logger = logging.getLogger("bayesflow")


def _log(msg, *args, callback_fn: callable = print, **kwargs):
    if keras.backend.backend() == "jax":
        import jax

        def __log(*a, **k):
            callback_fn(msg.format(*a, **k))

        jax.debug.callback(__log, *args, **kwargs)
    else:
        callback_fn(msg.format(*args, **kwargs))


def critical(msg, *args, **kwargs):
    _log(msg, *args, callback_fn=logger.critical, **kwargs)


def debug(msg, *args, **kwargs):
    _log(msg, *args, callback_fn=logger.debug, **kwargs)


def error(msg, *args, **kwargs):
    _log(msg, *args, callback_fn=logger.error, **kwargs)


def exception(msg, *args, **kwargs):
    _log(msg, *args, callback_fn=logger.exception, **kwargs)


def info(msg, *args, **kwargs):
    _log(msg, *args, callback_fn=logger.info, **kwargs)


def log(msg, *args, **kwargs):
    _log(msg, *args, callback_fn=logger.log, **kwargs)


def warning(msg, *args, **kwargs):
    _log(msg, *args, callback_fn=logger.warning, **kwargs)


@lru_cache(None)
def warn_once(msg, *args, **kwargs):
    warning(msg, *args, **kwargs)



from typing import Sequence, Mapping, Any

import numpy as np

import keras

from bayesflow.utils.exceptions import ShapeError
from bayesflow.networks import MLP


def classifier_two_sample_test(
    estimates: np.ndarray,
    targets: np.ndarray,
    metric: str = "accuracy",
    patience: int = 5,
    max_epochs: int = 1000,
    batch_size: int = 128,
    return_metric_only: bool = True,
    validation_split: float = 0.5,
    standardize: bool = True,
    mlp_widths: Sequence = (64, 64),
    **kwargs,
) -> float | Mapping[str, Any]:
    """
    C2ST metric [1] between samples from two distributions computed using a neural classifier.
    Can be computationally expensive if called in a loop[, since it needs to train the model
    for each set of samples.

    Note: works best for large numbers of samples and averaged across different posteriors.

    [1] Lopez-Paz, D., & Oquab, M. (2016). Revisiting classifier two-sample tests. arXiv:1610.06545.

    Parameters
    ----------
    estimates : np.ndarray
        Array of shape (num_samples_est, num_variables) containing samples representing estimated quantities
        (e.g., approximate posterior samples).
    targets : np.ndarray
        Array of shape (num_samples_tar, num_variables) containing target samples
        (e.g., samples from a reference posterior).
    metric : str, optional
        Metric to evaluate the classifier performance. Default is "accuracy".
    patience : int, optional
        Number of epochs with no improvement after which training will be stopped. Default is 5.
    max_epochs : int, optional
        Maximum number of epochs to train the classifier. Default is 1000.
    batch_size : int, optional
        Number of samples per batch during training. Default is 64.
    return_metric_only : bool, optional
        If True, only the final validation metric is returned. Otherwise, a dictionary with the score, classifier, and
        full training history is returned. Default is True.
    validation_split : float, optional
        Fraction of the training data to be used as validation data. Default is 0.5.
    standardize : bool, optional
        If True, both estimates and targets will be standardized using the mean and standard deviation of estimates.
        Default is True.
    mlp_widths : Sequence[int], optional
        Sequence specifying the number of units in each hidden layer of the MLP classifier. Default is (256, 256).
    **kwargs
        Additional keyword arguments. Recognized keyword:
            mlp_kwargs : dict
                Dictionary of additional parameters to pass to the MLP constructor.

    Returns
    -------
    results : float or dict
        If return_metric_only is True, returns the final validation metric (e.g., accuracy) as a float.
        Otherwise, returns a dictionary with keys "score", "classifier", and "history", where "score"
        is the final validation metric, "classifier" is the trained Keras model, and "history" contains the
        full training history.
    """

    # Error, if targets dim does not match estimates dim
    num_dims = estimates.shape[1]
    if not num_dims == targets.shape[1]:
        raise ShapeError(
            f"estimates and targets can have different number of samples (1st dim)"
            f"but must have the same dimensionality (2nd dim)"
            f"found: estimates shape {estimates.shape[1]}, targets shape {targets.shape[1]}"
        )

    # Standardize both estimates and targets relative to estimates mean and std
    if standardize:
        estimates_mean = np.mean(estimates, axis=0)
        estimates_std = np.std(estimates, axis=0)
        estimates = (estimates - estimates_mean) / estimates_std
        targets = (targets - estimates_mean) / estimates_std

    # Create data for classification task
    data = np.r_[estimates, targets]
    labels = np.r_[np.zeros((estimates.shape[0],)), np.ones((targets.shape[0],))]

    # Important: needed, since keras does not shuffle before selecting validation split
    shuffle_idx = np.random.permutation(data.shape[0])
    data = data[shuffle_idx]
    labels = labels[shuffle_idx]

    # Create and train classifier with optional stopping
    classifier = keras.Sequential(
        [MLP(widths=mlp_widths, **kwargs.get("mlp_kwargs", {})), keras.layers.Dense(1, activation="sigmoid")]
    )

    classifier.compile(optimizer="adam", loss="binary_crossentropy", metrics=[metric])

    early_stopping = keras.callbacks.EarlyStopping(
        monitor=f"val_{metric}", patience=patience, restore_best_weights=True
    )

    # For now, we need to enable grads, since we turn them off by default
    if keras.backend.backend() == "torch":
        import torch

        with torch.enable_grad():
            history = classifier.fit(
                x=data,
                y=labels,
                epochs=max_epochs,
                batch_size=batch_size,
                verbose=0,
                callbacks=[early_stopping],
                validation_split=validation_split,
            )
    else:
        history = classifier.fit(
            x=data,
            y=labels,
            epochs=max_epochs,
            batch_size=batch_size,
            verbose=0,
            callbacks=[early_stopping],
            validation_split=validation_split,
        )

    if return_metric_only:
        return history.history[f"val_{metric}"][-1]
    return {"score": history.history[f"val_{metric}"][-1], "classifier": classifier, "history": history.history}



import math

import keras
from keras.saving import register_keras_serializable as serializable

from bayesflow.types import Shape, Tensor
from bayesflow.utils import find_network, serialize_value_or_type, deserialize_value_or_type


@serializable(package="bayesflow.scores")
class ScoringRule:
    """Base class for scoring rules.

    Scoring rules evaluate the quality of statistical predictions based on the values that materialize
    when sampling from the true distribution. By minimizing an expected score, estimates with
    different properties can be obtained.

    To define a custom :py:class:`ScoringRule`, inherit from this class and overwrite the score method.
    For proper serialization, any new constructor arguments must be taken care of in a `get_config` method.

    Estimates are typically parameterized by projection heads consisting of a neural network component
    and a link to project into the correct output space.

    A :py:class:`ScoringRule` can score estimates consisting of multiple parts. See :py:class:`MultivariateNormalScore`
    for an example of a :py:class:`ParametricDistributionScore`. That score evaluates an estimated mean
    and covariance simultaneously.
    """

    NOT_TRANSFORMING_LIKE_VECTOR_WARNING = tuple()
    """
    This variable contains names of prediction heads that should lead to a warning when the adapter is applied
    in inverse direction to them.

    Prediction heads can output estimates in spaces other than the target distribution space.
    To such estimates the adapter cannot be straightforwardly applied in inverse direction,
    because the adapter is built to map vectors from the inference variable space. When subclassing
    :py:class:`ScoringRule`, add the names of such heads to the following list to warn users about difficulties
    with a type of estimate whenever the adapter is applied to them in inverse direction.
    """

    def __init__(
        self,
        subnets: dict[str, str | type] = None,
        subnets_kwargs: dict[str, dict] = None,
        links: dict[str, str | type] = None,
    ):
        self.subnets = subnets or {}
        self.subnets_kwargs = subnets_kwargs or {}
        self.links = links or {}

        self.config = {"subnets_kwargs": self.subnets_kwargs}

    def get_config(self):
        self.config["subnets"] = {
            key: serialize_value_or_type({}, "subnet", subnet) for key, subnet in self.subnets.items()
        }
        self.config["links"] = {key: serialize_value_or_type({}, "link", link) for key, link in self.links.items()}

        return self.config

    @classmethod
    def from_config(cls, config):
        config = config.copy()
        config["subnets"] = {
            key: deserialize_value_or_type(subnet_dict, "subnet")["subnet"]
            for key, subnet_dict in config["subnets"].items()
        }
        config["links"] = {
            key: deserialize_value_or_type(link_dict, "link")["link"] for key, link_dict in config["links"].items()
        }

        return cls(**config)

    def get_head_shapes_from_target_shape(self, target_shape: Shape) -> dict[str, Shape]:
        """Request a dictionary of names and output shapes of required heads from the score."""
        raise NotImplementedError

    def get_subnet(self, key: str) -> keras.Layer:
        """For a specified key, request a subnet to be used for projecting the shared condition embedding
        before further projection and reshaping to the heads output shape.

        If no subnet was specified for the key (e.g. upon initialization),
        return just an instance of keras.layers.Identity.

        Parameters
        ----------
        key : str
            Name of head for which to request a subnet.

        Returns
        -------
        link : keras.Layer
            Subnet projecting the shared condition embedding.
        """
        if key not in self.subnets.keys():
            return keras.layers.Identity()
        else:
            return find_network(self.subnets[key], **self.subnets_kwargs.get(key, {}))

    def get_link(self, key: str) -> keras.Layer:
        """For a specified key, request a link from network output to estimation target.

        If no link was specified for the key (e.g. upon initialization), return a linear activation.

        Parameters
        ----------
        key : str
            Name of head for which to request a link.

        Returns
        -------
        link : keras.Layer
            Activation function linking network output to estimation target.
        """
        if key not in self.links.keys():
            return keras.layers.Activation("linear")
        elif isinstance(self.links[key], str):
            return keras.layers.Activation(self.links[key])
        else:
            return self.links[key]

    def get_head(self, key: str, output_shape: Shape) -> keras.Sequential:
        """For a specified head key and output shape, request corresponding head network.

        A head network has the following components that are called sequentially:

        1. subnet: A keras.Layer.
        2. dense: A trainable linear projection with as many units as are required by the next component.
        3. reshape: Changes shape of output of projection to match requirements of next component.
        4. link: Transforms unconstrained values into a constrained space for the final estimator.
           See :py:mod:`~bayesflow.links` for examples.

        This method initializes the components in reverse order to meet all requirements and returns them.

        Parameters
        ----------
        key : str
            Name of head for which to request a link.
        output_shape: Shape
            The necessary shape of estimated values for the given key as returned by
            :py:func:`get_head_shapes_from_target_shape()`.

        Returns
        -------
        head : keras.Sequential
            Head network consisting of a learnable projection, a reshape and a link operation
            to parameterize estimates.
        """
        # initialize head components back to front
        link = self.get_link(key)

        # link input shape can differ from output shape
        if hasattr(link, "compute_input_shape"):
            link_input_shape = link.compute_input_shape(output_shape)
        else:
            link_input_shape = output_shape

        reshape = keras.layers.Reshape(target_shape=link_input_shape)
        dense = keras.layers.Dense(units=math.prod(link_input_shape))
        subnet = self.get_subnet(key)

        return keras.Sequential([subnet, dense, reshape, link])

    def score(self, estimates: dict[str, Tensor], targets: Tensor, weights: Tensor) -> Tensor:
        """Scores a batch of probabilistic estimates of distributions based on samples
        of the corresponding distributions.

        Parameters
        ----------
        estimates : dict[str, Tensor]
            Dictionary of estimates.
        targets : Tensor
            Array of samples from the true distribution to evaluate the estimates.
        weights : Tensor
            Array of weights for aggregating the scores.

        Returns
        -------
        numeric_score : Tensor
            Negatively oriented score evaluating the estimates, aggregated for the whole batch.

        Examples
        --------
        The following shows how to score estimates with a :py:class:`MeanScore`. All :py:class:`ScoringRule`\ s
        follow this pattern, only differing in the structure of the estimates dictionary.

        >>> import keras
        >>> from bayesflow.scores import MeanScore
        >>>
        >>> # batch of samples from a normal distribution
        >>> samples = keras.random.normal(shape=(100,))
        >>>
        >>> # batch of uninformed (random) estimates
        >>> bad_estimates = {"value": keras.random.uniform((100,))}
        >>>
        >>> # batch of estimates that are closer to the true mean
        >>> better_estimates = {"value": keras.random.normal(stddev=0.1, shape=(100,))}
        >>>
        >>> # calculate the score
        >>> scoring_rule = MeanScore()
        >>> scoring_rule.score(bad_estimates, samples)
        <tf.Tensor: shape=(), dtype=float32, numpy=1.2243813276290894>
        >>> scoring_rule.score(better_estimates, samples)
        <tf.Tensor: shape=(), dtype=float32, numpy=1.013983130455017>
        """
        raise NotImplementedError



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import os
import sys

try:
    from sphinx_polyversion import load
    from sphinx_polyversion.git import GitRef
    from sphinx_polyversion.api import LoadError

    USE_POLYVERSION = True
    data = load(globals())
    current = data["current"].name
except ImportError:
    USE_POLYVERSION = False
    print("sphinx_polyversion not installed, building single version")
    current = "local"

sys.path.insert(0, os.path.abspath("../.."))
sys.path.insert(0, os.path.abspath("sphinxext"))

# might set copyright end to wrong year -> remove
if "SOURCE_DATE_EPOCH" in os.environ:
    del os.environ["SOURCE_DATE_EPOCH"]

# -- Project information -----------------------------------------------------

project = "BayesFlow"
author = "The BayesFlow authors"
copyright = "2023-%Y, BayesFlow authors (lead maintainer: Stefan T. Radev)"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "numpydoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "myst_nb",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx_design",
    "sphinxcontrib.bibtex",
]

if not current.startswith("v1."):
    extensions.extend(
        [
            "override_pst_pagetoc",  # local, see sphinxext folder
            "adapt_autodoc_docstring",  # local, see sphinxext folder
        ]
    )

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "matplotlib": ("https://matplotlib.org/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "seaborn": ("https://seaborn.pydata.org/", None),
}

bibtex_bibfiles = ["references.bib"]

numpydoc_show_class_members = False
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
]
myst_url_schemes = ["http", "https", "mailto"]

# Define shorthand for external links:
extlinks = {
    "mainbranch": (f"https://github.com/bayesflow-org/bayesflow/blob/{current}/%s", None),
}

coverage_show_missing_items = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# Options for autodoc and autosummary
# do not ignore __all__, use it to determine public members
autosummary_ignore_module_all = False
autosummary_imported_members = False
# selects content to insert into the main body of an autoclass directive.
autoclass_content = "both"

if current.startswith("v1."):
    autodoc_default_options = {
        "members": True,
        "undoc-members": True,
        "imported-members": False,
        "inherited-members": True,
        "show-inheritance": True,
        "special-members": "__call__",
        "memberorder": "bysource",
    }

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
html_title = "BayesFlow: Amortized Bayesian Inference"

# Add any paths that contain custom _static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin _static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["custom.css"]
html_show_sourcelink = False
html_theme_options = {
    "use_edit_page_button": True,
    "logo": {
        "alt-text": "BayesFlow",
        "image_light": "_static/bayesflow_hor.png",
        "image_dark": "_static/bayesflow_hor_dark.png",
    },
    "icon_links_label": "Icon Links",
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/bayesflow-org/bayesflow",
            "icon": "fa-brands fa-square-github",
            "type": "fontawesome",
        },
        {
            "name": "Discourse Forum",
            "url": "https://discuss.bayesflow.org/",
            "icon": "fa-brands fa-discourse",
            "type": "fontawesome",
        },
    ],
    "navbar_align": "left",
    # -- Template placement in theme layouts ----------------------------------
    "navbar_start": ["navbar-logo"],
    # Note that the alignment of navbar_center is controlled by navbar_align
    "navbar_center": ["navbar-nav"],
    "navbar_end": ["theme-switcher", "navbar-icon-links", "version-switcher"],
    # navbar_persistent is persistent right (even when on mobiles)
    "navbar_persistent": ["search-button"],
    "switcher": {
        "json_url": "/versions.json",
        "version_match": current,
    },
    "check_switcher": False,
}
html_context = {
    "github_url": "https://github.com",  # or your GitHub Enterprise site
    "github_user": "bayesflow-org",
    "github_repo": "bayesflow",
    "github_version": current,
    "doc_path": "docsrc/source",
}
html_logo = "_static/bayesflow_hor.png"
html_favicon = "_static/bayesflow_hex.ico"
html_baseurl = "https://www.bayesflow.org/"

todo_include_todos = True

# do not execute jupyter notebooks when building docs
nb_execution_mode = "off"

# download notebooks as .ipynb and not as .ipynb.txt
html_sourcelink_suffix = ""

suppress_warnings = [
    f"autosectionlabel._examples/{filename.split('.')[0]}"
    for filename in os.listdir("../../examples")
    if os.path.isfile(os.path.join("../../examples", filename))
]  # Avoid duplicate label warnings for Jupyter notebooks.

remove_from_toctrees = ["_autosummary/*"]

autosummmary_generate = True

# versioning data for template
if USE_POLYVERSION:
    try:
        data = load(globals())
        current: GitRef = data["current"]
    except LoadError:
        print("sphinx_polyversion could not load. Building single version")



import keras
import pytest


def test_require_argument_k():
    from bayesflow.scores import NormedDifferenceScore

    with pytest.raises(TypeError) as excinfo:
        NormedDifferenceScore()

    assert "missing 1 required positional argument: 'k'" in str(excinfo)


def test_score_output(scoring_rule, random_conditions):
    if random_conditions is None:
        random_conditions = keras.ops.convert_to_tensor([[1.0, 1.0]])

    # Using random random_conditions also as targets for the purpose of this test.
    head_shapes = scoring_rule.get_head_shapes_from_target_shape(random_conditions.shape)
    print(scoring_rule.get_config())
    estimates = {}
    for key, output_shape in head_shapes.items():
        link = scoring_rule.get_link(key)
        if hasattr(link, "compute_input_shape"):
            link_input_shape = link.compute_input_shape(output_shape)
        else:
            link_input_shape = output_shape
        dummy_input = keras.random.normal((random_conditions.shape[0],) + link_input_shape)
        estimates[key] = link(dummy_input)

    score = scoring_rule.score(estimates, random_conditions)

    assert score.ndim == 0


def test_mean_score_optimality(mean_score, random_conditions):
    if random_conditions is None:
        random_conditions = keras.ops.convert_to_tensor([[1.0]])

    key = "value"
    suboptimal_estimates = {key: keras.random.uniform(random_conditions.shape)}
    optimal_estimates = {key: random_conditions}

    suboptimal_score = mean_score.score(suboptimal_estimates, random_conditions)
    optimal_score = mean_score.score(optimal_estimates, random_conditions)

    assert suboptimal_score > optimal_score
    assert keras.ops.isclose(optimal_score, 0)


def test_unconditional_mvn(multivariate_normal_score):
    mean = keras.ops.convert_to_tensor([[0.0, 1.0]])
    covariance = keras.ops.convert_to_tensor([[[1.0, 0.0], [0.0, 1.0]]])
    multivariate_normal_score.sample((10,), mean, covariance)


def test_unconditional_mvn_value_error(multivariate_normal_score):
    mean = keras.ops.convert_to_tensor([0.0, 1.0])
    covariance = keras.ops.convert_to_tensor([[1.0, 0.0], [0.0, 1.0]])

    with pytest.raises(ValueError):
        multivariate_normal_score.sample((10,), mean, covariance)



"""
Unstable or largely untested networks, proceed with caution.
"""

from .cif import CIF
from .continuous_time_consistency_model import ContinuousTimeConsistencyModel
from .free_form_flow import FreeFormFlow

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])



from functools import partial

import keras

from bayesflow.utils.serialization import serializable
from .functional import maximum_mean_discrepancy


@serializable
class MaximumMeanDiscrepancy(keras.Metric):
    def __init__(
        self,
        name: str = "maximum_mean_discrepancy",
        kernel: str = "inverse_multiquadratic",
        unbiased: bool = False,
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)
        self.mmd = self.add_variable(shape=(), initializer="zeros", name="mmd")
        self.mmd_fn = partial(maximum_mean_discrepancy, kernel=kernel, unbiased=unbiased)

    def update_state(self, x, y):
        self.mmd.assign(keras.ops.cast(self.mmd_fn(x, y), self.dtype))

    def result(self):
        return self.mmd.value



from collections.abc import Callable, MutableSequence, Sequence, Mapping

import numpy as np

from bayesflow.utils.serialization import deserialize, serialize, serializable

from .transforms import (
    AsSet,
    AsTimeSeries,
    Broadcast,
    Concatenate,
    Constrain,
    ConvertDType,
    Drop,
    ExpandDims,
    FilterTransform,
    Keep,
    Log,
    MapTransform,
    NumpyTransform,
    OneHot,
    Rename,
    SerializableCustomTransform,
    Sqrt,
    Standardize,
    ToArray,
    Transform,
)
from .transforms.filter_transform import Predicate


@serializable
class Adapter(MutableSequence[Transform]):
    """
    Defines an adapter to apply various transforms to data.

    Where possible, the transforms also supply an inverse transform.

    Parameters
    ----------
    transforms : Sequence[Transform], optional
        The sequence of transforms to execute.
    """

    def __init__(self, transforms: Sequence[Transform] | None = None):
        if transforms is None:
            transforms = []

        self.transforms = list(transforms)

    @staticmethod
    def create_default(inference_variables: Sequence[str]) -> "Adapter":
        """Create an adapter with a set of default transforms.

        Parameters
        ----------
        inference_variables : Sequence of str
            The names of the variables to be inferred by an estimator.

        Returns
        -------
        An initialized Adapter with a set of default transforms.
        """
        return (
            Adapter()
            .to_array()
            .convert_dtype("float64", "float32")
            .concatenate(inference_variables, into="inference_variables")
        )

    @classmethod
    def from_config(cls, config: dict, custom_objects=None) -> "Adapter":
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self) -> dict:
        config = {
            "transforms": self.transforms,
        }

        return serialize(config)

    def forward(self, data: dict[str, any], *, stage: str = "inference", **kwargs) -> dict[str, np.ndarray]:
        """Apply the transforms in the forward direction.

        Parameters
        ----------
        data : dict
            The data to be transformed.
        stage : str, one of ["training", "validation", "inference"]
            The stage the function is called in.
        **kwargs : dict
            Additional keyword arguments passed to each transform.

        Returns
        -------
        dict
            The transformed data.
        """
        data = data.copy()

        for transform in self.transforms:
            data = transform(data, stage=stage, **kwargs)

        return data

    def inverse(self, data: dict[str, np.ndarray], *, stage: str = "inference", **kwargs) -> dict[str, any]:
        """Apply the transforms in the inverse direction.

        Parameters
        ----------
        data : dict
            The data to be transformed.
        stage : str, one of ["training", "validation", "inference"]
            The stage the function is called in.
        **kwargs : dict
            Additional keyword arguments passed to each transform.

        Returns
        -------
        dict
            The transformed data.
        """
        data = data.copy()

        for transform in reversed(self.transforms):
            data = transform(data, stage=stage, inverse=True, **kwargs)

        return data

    def __call__(
        self, data: Mapping[str, any], *, inverse: bool = False, stage="inference", **kwargs
    ) -> dict[str, np.ndarray]:
        """Apply the transforms in the given direction.

        Parameters
        ----------
        data : Mapping[str, any]
            The data to be transformed.
        inverse : bool, optional
            If False, apply the forward transform, else apply the inverse transform (default False).
        stage : str, one of ["training", "validation", "inference"]
            The stage the function is called in.
        **kwargs
            Additional keyword arguments passed to each transform.

        Returns
        -------
        dict
            The transformed data.
        """
        if inverse:
            return self.inverse(data, stage=stage, **kwargs)

        return self.forward(data, stage=stage, **kwargs)

    def __repr__(self):
        result = ""
        for i, transform in enumerate(self):
            result += f"{i}: {transform!r}"
            if i != len(self) - 1:
                result += " -> "

        return f"Adapter([{result}])"

    # list methods

    def append(self, value: Transform) -> "Adapter":
        """Append a transform to the list of transforms.

        Parameters
        ----------
        value : Transform
            The transform to be added.
        """
        self.transforms.append(value)
        return self

    def __delitem__(self, key: int | slice):
        del self.transforms[key]

    def extend(self, values: Sequence[Transform]) -> "Adapter":
        """Extend the adapter with a sequence of transforms.

        Parameters
        ----------
        values : Sequence of Transform
            The additional transforms to extend the adapter.
        """
        if isinstance(values, Adapter):
            values = values.transforms

        self.transforms.extend(values)

        return self

    def __getitem__(self, item: int | slice) -> "Adapter":
        if isinstance(item, int):
            return self.transforms[item]

        return Adapter(self.transforms[item])

    def insert(self, index: int, value: Transform | Sequence[Transform]) -> "Adapter":
        """Insert a transform at a given index.

        Parameters
        ----------
        index : int
            The index to insert at.
        value : Transform or Sequence of Transform
            The transform or transforms to insert.
        """
        if isinstance(value, Adapter):
            value = value.transforms

        if isinstance(value, Sequence):
            # convenience: Adapters are always flat
            self.transforms = self.transforms[:index] + list(value) + self.transforms[index:]
        else:
            self.transforms.insert(index, value)

        return self

    def __setitem__(self, key: int | slice, value: Transform | Sequence[Transform]) -> "Adapter":
        if isinstance(value, Adapter):
            value = value.transforms

        if isinstance(key, int) and isinstance(value, Sequence):
            if key < 0:
                key += len(self.transforms)

            key = slice(key, key + 1)

        self.transforms[key] = value

        return self

    def __len__(self):
        return len(self.transforms)

    # adapter methods

    add_transform = append

    def apply(
        self,
        include: str | Sequence[str] = None,
        *,
        forward: np.ufunc | str,
        inverse: np.ufunc | str = None,
        predicate: Predicate = None,
        exclude: str | Sequence[str] = None,
        **kwargs,
    ):
        """Append a :py:class:`~transforms.NumpyTransform` to the adapter.

        Parameters
        ----------
        forward : str or np.ufunc
            The name of the NumPy function to use for the forward transformation.
        inverse : str or np.ufunc, optional
            The name of the NumPy function to use for the inverse transformation.
            By default, the inverse is inferred from the forward argument for supported methods.
            You can find the supported methods in
            :py:const:`~bayesflow.adapters.transforms.NumpyTransform.INVERSE_METHODS`.
        predicate : Predicate, optional
            Function that indicates which variables should be transformed.
        include : str or Sequence of str, optional
            Names of variables to include in the transform.
        exclude : str or Sequence of str, optional
            Names of variables to exclude from the transform.
        **kwargs : dict
            Additional keyword arguments passed to the transform.
        """
        transform = FilterTransform(
            transform_constructor=NumpyTransform,
            predicate=predicate,
            include=include,
            exclude=exclude,
            forward=forward,
            inverse=inverse,
            **kwargs,
        )
        self.transforms.append(transform)
        return self

    def apply_serializable(
        self,
        include: str | Sequence[str] = None,
        *,
        forward: Callable[[np.ndarray, ...], np.ndarray],
        inverse: Callable[[np.ndarray, ...], np.ndarray],
        predicate: Predicate = None,
        exclude: str | Sequence[str] = None,
        **kwargs,
    ):
        """Append a :py:class:`~transforms.SerializableCustomTransform` to the adapter.

        Parameters
        ----------
        forward : function, no lambda
            Registered serializable function to transform the data in the forward pass.
            For the adapter to be serializable, this function has to be serializable
            as well (see Notes). Therefore, only proper functions and no lambda
            functions can be used here.
        inverse : function, no lambda
            Registered serializable function to transform the data in the inverse pass.
            For the adapter to be serializable, this function has to be serializable
            as well (see Notes). Therefore, only proper functions and no lambda
            functions can be used here.
        predicate : Predicate, optional
            Function that indicates which variables should be transformed.
        include : str or Sequence of str, optional
            Names of variables to include in the transform.
        exclude : str or Sequence of str, optional
            Names of variables to exclude from the transform.
        **kwargs : dict
            Additional keyword arguments passed to the transform.

        Raises
        ------
        ValueError
            When the provided functions are not registered serializable functions.

        Notes
        -----
        Important: The forward and inverse functions have to be registered with Keras.
        To do so, use the `@keras.saving.register_keras_serializable` decorator.
        They must also be registered (and identical) when loading the adapter
        at a later point in time.

        Examples
        --------

        The example below shows how to use the
        `keras.saving.register_keras_serializable` decorator to
        register functions with Keras. Note that for this simple
        example, one usually would use the simpler :py:meth:`apply`
        method.

        >>> import keras
        >>>
        >>> @keras.saving.register_keras_serializable("custom")
        >>> def forward_fn(x):
        >>>     return x**2
        >>>
        >>> @keras.saving.register_keras_serializable("custom")
        >>> def inverse_fn(x):
        >>>     return x**0.5
        >>>
        >>> adapter = bf.Adapter().apply_serializable(
        >>>     "x",
        >>>     forward=forward_fn,
        >>>     inverse=inverse_fn,
        >>> )
        """
        transform = FilterTransform(
            transform_constructor=SerializableCustomTransform,
            predicate=predicate,
            include=include,
            exclude=exclude,
            forward=forward,
            inverse=inverse,
            **kwargs,
        )
        self.transforms.append(transform)
        return self

    def as_set(self, keys: str | Sequence[str]):
        """Append an :py:class:`~transforms.AsSet` transform to the adapter.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to apply the transform to.
        """
        if isinstance(keys, str):
            keys = [keys]

        transform = MapTransform({key: AsSet() for key in keys})
        self.transforms.append(transform)
        return self

    def as_time_series(self, keys: str | Sequence[str]):
        """Append an :py:class:`~transforms.AsTimeSeries` transform to the adapter.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to apply the transform to.
        """
        if isinstance(keys, str):
            keys = [keys]

        transform = MapTransform({key: AsTimeSeries() for key in keys})
        self.transforms.append(transform)
        return self

    def broadcast(
        self,
        keys: str | Sequence[str],
        *,
        to: str,
        expand: str | int | tuple = "left",
        exclude: int | tuple = -1,
        squeeze: int | tuple = None,
    ):
        """Append a :py:class:`~transforms.Broadcast` transform to the adapter.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to apply the transform to.
        to : str
            Name of the data variable to broadcast to.
        expand : str or int or tuple, optional
            Where should new dimensions be added to match the number of dimensions in `to`?
            Can be "left", "right", or an integer or tuple containing the indices of the new dimensions.
            The latter is needed if we want to include a dimension in the middle, which will be required
            for more advanced cases. By default we expand left.
        exclude : int or tuple, optional
            Which dimensions (of the dimensions after expansion) should retain their size,
            rather than being broadcasted to the corresponding dimension size of `to`?
            By default we exclude the last dimension (usually the data dimension) from broadcasting the size.
        squeeze : int or tuple, optional
            Axis to squeeze after broadcasting.

        Notes
        -----
        Important: Do not broadcast to variables that are used as inference variables
        (i.e., parameters to be inferred by the networks). The adapter will work during training
        but then fail during inference because the variable being broadcasted to is not available.
        """
        if isinstance(keys, str):
            keys = [keys]

        transform = Broadcast(keys, to=to, expand=expand, exclude=exclude, squeeze=squeeze)
        self.transforms.append(transform)
        return self

    def clear(self):
        """Remove all transforms from the adapter."""
        self.transforms = []
        return self

    def concatenate(self, keys: str | Sequence[str], *, into: str, axis: int = -1):
        """Append a :py:class:`~transforms.Concatenate` transform to the adapter.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to concatenate.
        into : str
            The name of the resulting variable.
        axis : int, optional
            Along which axis to concatenate the keys. The last axis is used by default.
        """
        if isinstance(keys, str):
            transform = Rename(keys, to_key=into)
        else:
            transform = Concatenate(keys, into=into, axis=axis)
        self.transforms.append(transform)
        return self

    def convert_dtype(
        self,
        from_dtype: str,
        to_dtype: str,
        *,
        predicate: Predicate = None,
        include: str | Sequence[str] = None,
        exclude: str | Sequence[str] = None,
    ):
        """Append a :py:class:`~transforms.ConvertDType` transform to the adapter.
        See also :py:meth:`~bayesflow.adapters.Adapter.map_dtype`.

        Parameters
        ----------
        from_dtype : str
            Original dtype
        to_dtype : str
            Target dtype
        predicate : Predicate, optional
            Function that indicates which variables should be transformed.
        include : str or Sequence of str, optional
            Names of variables to include in the transform.
        exclude : str or Sequence of str, optional
            Names of variables to exclude from the transform.
        """
        transform = FilterTransform(
            transform_constructor=ConvertDType,
            predicate=predicate,
            include=include,
            exclude=exclude,
            from_dtype=from_dtype,
            to_dtype=to_dtype,
        )
        self.transforms.append(transform)
        return self

    def constrain(
        self,
        keys: str | Sequence[str],
        *,
        lower: int | float | np.ndarray = None,
        upper: int | float | np.ndarray = None,
        method: str = "default",
        inclusive: str = "both",
        epsilon: float = 1e-15,
    ):
        """Append a :py:class:`~transforms.Constrain` transform to the adapter.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to constrain.
        lower: int or float or np.darray, optional
            Lower bound for named data variable.
        upper : int or float or np.darray, optional
            Upper bound for named data variable.
        method : str, optional
            Method by which to shrink the network predictions space to specified bounds. Choose from
            - Double bounded methods: sigmoid, expit, (default = sigmoid)
            - Lower bound only methods: softplus, exp, (default = softplus)
            - Upper bound only methods: softplus, exp, (default = softplus)
        inclusive : {'both', 'lower', 'upper', 'none'}, optional
            Indicates which bounds are inclusive (or exclusive).
            - "both" (default): Both lower and upper bounds are inclusive.
            - "lower": Lower bound is inclusive, upper bound is exclusive.
            - "upper": Lower bound is exclusive, upper bound is inclusive.
            - "none": Both lower and upper bounds are exclusive.
        epsilon : float, optional
            Small value to ensure inclusive bounds are not violated.
            Current default is 1e-15 as this ensures finite outcomes
            with the default transformations applied to data exactly at the boundaries.
        """
        if isinstance(keys, str):
            keys = [keys]

        transform = MapTransform(
            transform_map={
                key: Constrain(lower=lower, upper=upper, method=method, inclusive=inclusive, epsilon=epsilon)
                for key in keys
            }
        )
        self.transforms.append(transform)
        return self

    def drop(self, keys: str | Sequence[str]):
        """Append a :py:class:`~transforms.Drop` transform to the adapter.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to drop.
        """
        if isinstance(keys, str):
            keys = [keys]

        transform = Drop(keys)
        self.transforms.append(transform)
        return self

    def expand_dims(self, keys: str | Sequence[str], *, axis: int | tuple):
        """Append an :py:class:`~transforms.ExpandDims` transform to the adapter.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to expand.
        axis : int or tuple
            The axis to expand.
        """
        if isinstance(keys, str):
            keys = [keys]

        transform = MapTransform({key: ExpandDims(axis=axis) for key in keys})
        self.transforms.append(transform)
        return self

    def keep(self, keys: str | Sequence[str]):
        """Append a :py:class:`~transforms.Keep` transform to the adapter.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to keep.
        """
        if isinstance(keys, str):
            keys = [keys]

        transform = Keep(keys)
        self.transforms.append(transform)
        return self

    def log(self, keys: str | Sequence[str], *, p1: bool = False):
        """Append an :py:class:`~transforms.Log` transform to the adapter.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to transform.
        p1 : boolean
            Add 1 to the input before taking the logarithm?
        """
        if isinstance(keys, str):
            keys = [keys]

        transform = MapTransform({key: Log(p1=p1) for key in keys})
        self.transforms.append(transform)
        return self

    def map_dtype(self, keys: str | Sequence[str], to_dtype: str):
        """Append a :py:class:`~transforms.ConvertDType` transform to the adapter.
        See also :py:meth:`~bayesflow.adapters.Adapter.convert_dtype`.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to transform.
        to_dtype : str
            Target dtype
        """
        if isinstance(keys, str):
            keys = [keys]

        transform = MapTransform({key: ConvertDType(to_dtype) for key in keys})
        self.transforms.append(transform)
        return self

    def one_hot(self, keys: str | Sequence[str], num_classes: int):
        """Append a :py:class:`~transforms.OneHot` transform to the adapter.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to transform.
        num_classes : int
            Number of classes for the encoding.
        """
        if isinstance(keys, str):
            keys = [keys]

        transform = MapTransform({key: OneHot(num_classes=num_classes) for key in keys})
        self.transforms.append(transform)
        return self

    def rename(self, from_key: str, to_key: str):
        """Append a :py:class:`~transforms.Rename` transform to the adapter.

        Parameters
        ----------
        from_key : str
            Variable name that should be renamed
        to_key : str
            New variable name
        """
        self.transforms.append(Rename(from_key, to_key))
        return self

    def scale(self, keys: str | Sequence[str], by: float | np.ndarray):
        from .transforms import Scale

        if isinstance(keys, str):
            keys = [keys]

        self.transforms.append(MapTransform({key: Scale(scale=by) for key in keys}))
        return self

    def shift(self, keys: str | Sequence[str], by: float | np.ndarray):
        from .transforms import Shift

        if isinstance(keys, str):
            keys = [keys]

        self.transforms.append(MapTransform({key: Shift(shift=by) for key in keys}))
        return self

    def split(self, key: str, *, into: Sequence[str], indices_or_sections: int | Sequence[int] = None, axis: int = -1):
        from .transforms import Split

        if isinstance(into, str):
            transform = Rename(key, into)
        else:
            transform = Split(key, into, indices_or_sections, axis)

        self.transforms.append(transform)

        return self

    def sqrt(self, keys: str | Sequence[str]):
        """Append an :py:class:`~transforms.Sqrt` transform to the adapter.

        Parameters
        ----------
        keys : str or Sequence of str
            The names of the variables to transform.
        """
        if isinstance(keys, str):
            keys = [keys]

        transform = MapTransform({key: Sqrt() for key in keys})
        self.transforms.append(transform)
        return self

    def standardize(
        self,
        include: str | Sequence[str] = None,
        *,
        predicate: Predicate = None,
        exclude: str | Sequence[str] = None,
        **kwargs,
    ):
        """Append a :py:class:`~transforms.Standardize` transform to the adapter.

        Parameters
        ----------
        predicate : Predicate, optional
            Function that indicates which variables should be transformed.
        include : str or Sequence of str, optional
            Names of variables to include in the transform.
        exclude : str or Sequence of str, optional
            Names of variables to exclude from the transform.
        **kwargs : dict
            Additional keyword arguments passed to the transform.
        """
        transform = FilterTransform(
            transform_constructor=Standardize,
            predicate=predicate,
            include=include,
            exclude=exclude,
            **kwargs,
        )
        self.transforms.append(transform)
        return self

    def to_array(
        self,
        include: str | Sequence[str] = None,
        *,
        predicate: Predicate = None,
        exclude: str | Sequence[str] = None,
        **kwargs,
    ):
        """Append a :py:class:`~transforms.ToArray` transform to the adapter.

        Parameters
        ----------
        predicate : Predicate, optional
            Function that indicates which variables should be transformed.
        include : str or Sequence of str, optional
            Names of variables to include in the transform.
        exclude : str or Sequence of str, optional
            Names of variables to exclude from the transform.
        **kwargs : dict
            Additional keyword arguments passed to the transform.
        """
        transform = FilterTransform(
            transform_constructor=ToArray,
            predicate=predicate,
            include=include,
            exclude=exclude,
            **kwargs,
        )
        self.transforms.append(transform)
        return self

    def to_dict(self):
        from .transforms import ToDict

        transform = ToDict()
        self.transforms.append(transform)
        return self



from collections.abc import Sequence

from bayesflow.utils.serialization import serializable, serialize

from .transform import Transform


@serializable
class Keep(Transform):
    """
    Name the data parameters that should be kept for futher calculation.

    Parameters
    ----------
    keys : sequence of str
        The names of kept data variables as strings.

    Examples
    --------

    Two moons simulator generates data for priors alpha, r and theta as well as observation data x.
    We are interested only in theta and x, to keep only theta and x we should use the following;

    >>> adapter = (
        bf.adapters.Adapter()
            # drop data from unneeded priors alpha, and r
            # only keep theta and x
            .keep(("theta", "x"))
        )

    The following example shows the usage in a more general case:

    >>> a = [1, 2, 3, 4]
    >>> b = [[1, 2], [3, 4]]
    >>> c = [[5, 6, 7, 8]]
    >>> dat = dict(a=a, b=b, c=c)

    Here we want to only keep elements b and c

    >>> keeper = bf.adapters.transforms.Keep(("b", "c"))
    >>> keeper.forward(dat)
    {'b': [[1, 2], [3, 4]], 'c': [[5, 6, 7, 8]]}
    """

    def __init__(self, keys: Sequence[str]):
        self.keys = keys

    def get_config(self) -> dict:
        return serialize({"keys": self.keys})

    def forward(self, data: dict[str, any], **kwargs) -> dict[str, any]:
        return {key: value for key, value in data.items() if key in self.keys}

    def inverse(self, data: dict[str, any], **kwargs) -> dict[str, any]:
        # non-invertible transform
        return data

    def extra_repr(self) -> str:
        return "[" + ", ".join(map(repr, self.keys)) + "]"



from typing import Sequence, Any, Mapping

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from matplotlib.collections import LineCollection
from matplotlib.colors import Normalize
from matplotlib.patches import Rectangle
from matplotlib.legend_handler import HandlerPatch

from .validators import check_estimates_prior_shapes
from .dict_utils import dicts_to_arrays


def prepare_plot_data(
    estimates: Mapping[str, np.ndarray] | np.ndarray,
    targets: Mapping[str, np.ndarray] | np.ndarray,
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    num_col: int = None,
    num_row: int = None,
    figsize: tuple = None,
    stacked: bool = False,
    default_name: str = "v",
) -> Mapping[str, Any]:
    """
    Procedural wrapper that encompasses all preprocessing steps, including shape-checking, parameter name
    generation, layout configuration, figure initialization, and collapsing of axes.

    Parameters
    ----------
    estimates           : dict[str, ndarray] or ndarray
        The model-generated predictions or estimates, which can take the following forms:

        - ndarray of shape (num_datasets, num_variables)
            Point estimates for each dataset, where `num_datasets` is the number of datasets
            and `num_variables` is the number of variables per dataset.
        - ndarray of shape (num_datasets, num_draws, num_variables)
            Posterior samples for each dataset, where `num_datasets` is the number of datasets,
            `num_draws` is the number of posterior draws, and `num_variables` is the number of variables.
    targets        : dict[str, ndarray] or ndarray, optional (default = None)
        Ground truth values corresponding to the estimates. Must match the structure and dimensionality
        of `estimates` in terms of first and last axis.
    variable_keys     : list or None, optional, default: None
       Select keys from the dictionary provided in samples. By default, select all keys.
    variable_names    : Sequence[str], optional (default = None)
        Optional variable names to act as a filter if dicts provided or actual variable names in case of array args
    num_col           : int
        Number of columns for the visualization layout
    num_row           : int
        Number of rows for the visualization layout
    figsize           : tuple, optional, default: None
        Size of the figure adjusting to the display resolution
    stacked           : bool, optional, default: False
        Whether the plots are stacked horizontally
    default_name      : str, optional (default = "v")
        The default name to use for estimates if None provided
    """

    plot_data = dicts_to_arrays(
        estimates=estimates,
        targets=targets,
        variable_keys=variable_keys,
        variable_names=variable_names,
        default_name=default_name,
    )
    check_estimates_prior_shapes(plot_data["estimates"], plot_data["targets"])

    # store variable information at top level for easy access
    variable_names = plot_data["estimates"].variable_names
    num_variables = len(variable_names)
    plot_data["variable_names"] = variable_names
    plot_data["num_variables"] = num_variables

    # Configure layout
    num_row, num_col = set_layout(num_variables, num_row, num_col, stacked)

    # Initialize figure
    fig, axes = make_figure(num_row, num_col, figsize=figsize)

    plot_data["fig"] = fig
    plot_data["axes"] = axes
    plot_data["num_row"] = num_row
    plot_data["num_col"] = num_col

    return plot_data


def set_layout(num_total: int, num_row: int = None, num_col: int = None, stacked: bool = False):
    """
    Determine the number of rows and columns in diagnostics visualizations.

    Parameters
    ----------
    num_total     : int
        Total number of parameters
    num_row       : int, default = None
        Number of rows for the visualization layout
    num_col       : int, default = None
        Number of columns for the visualization layout
    stacked     : bool, default = False
        Boolean that determines whether to stack the plot or not.

    Returns
    -------
    num_row       : int
        Number of rows for the visualization layout
    num_col       : int
        Number of columns for the visualization layout
    """
    if stacked:
        num_row, num_col = 1, 1
    else:
        if num_row is None and num_col is None:
            num_row = int(np.ceil(num_total / 6))
            num_col = int(np.ceil(num_total / num_row))
        elif num_row is None and num_col is not None:
            num_row = int(np.ceil(num_total / num_col))
        elif num_row is not None and num_col is None:
            num_col = int(np.ceil(num_total / num_row))

    return num_row, num_col


def make_figure(num_row: int = None, num_col: int = None, figsize: tuple = None):
    """
    Initialize a set of figures

    Parameters
    ----------
    num_row       : int
        Number of rows in a figure
    num_col       : int
        Number of columns in a figure
    figsize       : tuple
        Size of the figure adjusting to the display resolution
        or the user's choice

    Returns
    -------
    f, axes
        Initialized figures
    """
    if num_row == 1 and num_col == 1:
        f, axes = plt.subplots(1, 1, figsize=figsize)
    else:
        if figsize is None:
            figsize = (int(5 * num_col), int(5 * num_row))

        f, axes = plt.subplots(num_row, num_col, figsize=figsize)
    axes = np.atleast_1d(axes)

    return f, axes


def add_metric(
    ax,
    metric_text: str = None,
    metric_value: float = None,
    position: tuple = (0.1, 0.9),
    metric_fontsize: int = 12,
):
    """TODO: docstring"""
    if metric_text is None or metric_value is None:
        raise ValueError("Metric text and values must be provided to be add this metric.")

    metric_label = f"{metric_text} = {metric_value:.3f}"

    ax.text(
        position[0],
        position[1],
        metric_label,
        ha="left",
        va="center",
        transform=ax.transAxes,
        size=metric_fontsize,
    )


def add_x_labels(
    axes: np.ndarray,
    num_row: int = None,
    num_col: int = None,
    xlabel: Sequence[str] | str = None,
    label_fontsize: int = None,
):
    """TODO: docstring"""
    if num_row == 1:
        bottom_row = axes
    else:
        bottom_row = axes[num_row - 1, :] if num_col > 1 else axes
    for i, ax in enumerate(bottom_row):
        # If labels are in sequence, set them sequentially. Otherwise, one label fits all.
        ax.set_xlabel(xlabel if isinstance(xlabel, str) else xlabel[i], fontsize=label_fontsize)


def add_y_labels(axes: np.ndarray, num_row: int = None, ylabel: Sequence[str] | str = None, label_fontsize: int = None):
    """TODO: docstring"""

    if num_row == 1:  # if there is only one row, the ax array is 1D
        axes[0].set_ylabel(ylabel, fontsize=label_fontsize)
    # If there is more than one row, the ax array is 2D
    else:
        for i, ax in enumerate(axes[:, 0]):
            # If labels are in sequence, set them sequentially. Otherwise, one label fits all.
            ax.set_ylabel(ylabel if isinstance(ylabel, str) else ylabel[i], fontsize=label_fontsize)


def add_titles(axes: np.ndarray, title: Sequence[str] | str = None, title_fontsize: int = None):
    for t, ax in zip(title, axes.flat):
        ax.set_title(t, fontsize=title_fontsize)


def add_titles_and_labels(
    axes: np.ndarray,
    num_row: int = None,
    num_col: int = None,
    title: Sequence[str] | str = None,
    xlabel: Sequence[str] | str = None,
    ylabel: Sequence[str] | str = None,
    title_fontsize: int = None,
    label_fontsize: int = None,
):
    """
    Wrapper function for configuring labels for both axes.
    """
    if title is not None:
        add_titles(axes, title, title_fontsize)
    if xlabel is not None:
        add_x_labels(axes, num_row, num_col, xlabel, label_fontsize)
    if ylabel is not None:
        add_y_labels(axes, num_row, ylabel, label_fontsize)


def prettify_subplots(axes: np.ndarray, num_subplots: int, tick: bool = True, tick_fontsize: int = 12):
    """TODO: docstring"""
    for ax in axes.flat:
        sns.despine(ax=ax)
        ax.grid(alpha=0.5)
        if tick:
            ax.tick_params(axis="both", which="major", labelsize=tick_fontsize)
            ax.tick_params(axis="both", which="minor", labelsize=tick_fontsize)

    # Remove unused axes entirely
    for _ax in axes.flat[num_subplots:]:
        _ax.remove()


def make_quadratic(ax: plt.Axes, x_data: np.ndarray, y_data: np.ndarray):
    """
    Utility to make a subplots quadratic in order to avoid visual illusions
    in, e.g., recovery plots.
    """

    lower = min(x_data.min(), y_data.min())
    upper = max(x_data.max(), y_data.max())
    eps = (upper - lower) * 0.1
    ax.set_xlim((lower - eps, upper + eps))
    ax.set_ylim((lower - eps, upper + eps))
    ax.plot(
        [ax.get_xlim()[0], ax.get_xlim()[1]],
        [ax.get_ylim()[0], ax.get_ylim()[1]],
        color="black",
        alpha=0.9,
        linestyle="dashed",
    )


def gradient_line(x, y, c=None, cmap: str = "viridis", lw: float = 2.0, alpha: float = 1, ax=None):
    """
    Plot a 1D line with color gradient determined by `c` (same shape as x and y).
    """
    if ax is None:
        ax = plt.gca()

    # Default color value = y
    if c is None:
        c = y

    # Create segments for LineCollection
    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)

    norm = Normalize(np.min(c), np.max(c))
    lc = LineCollection(segments, array=c, cmap=cmap, norm=norm, linewidth=lw, alpha=alpha)

    ax.add_collection(lc)
    ax.set_xlim(np.min(x), np.max(x))
    ax.set_ylim(np.min(y), np.max(y))
    return lc


def gradient_legend(ax, label, cmap, norm, loc="upper right"):
    """
    Adds a single gradient swatch to the legend of the given Axes.

    Parameters
    ----------
    - ax: matplotlib Axes
    - label: str, label to display in the legend
    - cmap: matplotlib colormap
    - norm: matplotlib Normalize object
    - loc: legend location (default 'upper right')
    """

    # Custom dummy handle to represent the gradient
    class _GradientSwatch(Rectangle):
        pass

    # Custom legend handler that draws a horizontal gradient
    class _HandlerGradient(HandlerPatch):
        def create_artists(self, legend, orig_handle, xdescent, ydescent, width, height, fontsize, trans):
            gradient = np.linspace(0, 1, 256).reshape(1, -1)
            im = ax.imshow(
                gradient,
                aspect="auto",
                extent=[xdescent, xdescent + width, ydescent, ydescent + height],
                transform=trans,
                cmap=cmap,
                norm=norm,
            )
            return [im]

    # Add to existing legend entries
    handles, labels = ax.get_legend_handles_labels()
    handles.append(_GradientSwatch((0, 0), 1, 1))
    labels.append(label)

    ax.legend(handles=handles, labels=labels, loc=loc, handler_map={_GradientSwatch: _HandlerGradient()})


def add_gradient_plot(
    x,
    y,
    ax,
    cmap: str = "viridis",
    lw: float = 3.0,
    marker: bool = True,
    marker_type: str = "o",
    marker_size: int = 34,
    alpha: float = 1,
    label: str = "Validation",
):
    gradient_line(x, y, c=x, cmap=cmap, lw=lw, alpha=alpha, ax=ax)

    # Optionally add markers
    if marker:
        ax.scatter(
            x,
            y,
            c=x,
            cmap=cmap,
            marker=marker_type,
            s=marker_size,
            zorder=10,
            edgecolors="none",
            label=label,
            alpha=0.01,
        )
